import os

import aioboto3
from dasida import get_secrets
from dotenv import load_dotenv

load_dotenv()

# consts
PROD = "prod"
DEV = "dev"
SEATABLE_VIEW_SUFFIX_TO_WATCH = "__sync"

# AWS S3
AWS_S3_BUCKET_NAME = os.getenv("AWS_S3_BUCKET_NAME")
AWS_S3_BUCKET_PREFIX = os.getenv("AWS_S3_BUCKET_PREFIX")

# From Dasida
SM_AWS_S3 = os.getenv("SM_AWS_S3")
if SM_AWS_S3:
    AIOBOTO3_CONF = get_secrets(
        SM_AWS_S3,
        aws_access_key_id=os.getenv("AWS_ACCESS_KEY_ID"),
        aws_secret_access_key=os.getenv("AWS_SECRET_ACCESS_KEY"),
    )
else:
    AIOBOTO3_CONF = None



import asyncio
import io
import logging
import time
from typing import Union

import aioboto3
import orjson
import pyarrow as pa
import pyarrow.parquet as pq
import redis.asyncio as redis

from plantable.client import AdminClient, BaseClient
from plantable.model import Base, View

from .conf import (
    AWS_S3_ACCESS_KEY_ID,
    AWS_S3_BUCKET_NAME,
    AWS_S3_BUCKET_PREFIX,
    AWS_S3_REGION_NAME,
    AWS_S3_SECRET_ACCESS_KEY,
    KEY_PREFIX,
    REDIS_HOST,
    REDIS_PORT,
    SEATABLE_PASSWORD,
    SEATABLE_URL,
    SEATABLE_USERNAME,
)
from .events import event_parser
from .events.const import (
    COLUMN_EVENTS,
    OP_APPEND_ROW,
    OP_APPEND_ROWS,
    OP_ARCHIVE_ROW,
    OP_ARCHIVE_ROWS,
    OP_DELETE_COLUMN,
    OP_DELETE_ROW,
    OP_DELETE_ROWS,
    OP_DELETE_TABLE,
    OP_DELETE_VIEW,
    OP_INSERT_COLUMN,
    OP_INSERT_ROW,
    OP_INSERT_ROWS,
    OP_INSERT_TABLE,
    OP_INSERT_VIEW,
    OP_MODIFY_COLUMN_METADATA_PERMISSION,
    OP_MODIFY_COLUMN_PERMISSION,
    OP_MODIFY_COLUMN_TYPE,
    OP_MODIFY_FILTERS,
    OP_MODIFY_GROUPBYS,
    OP_MODIFY_HEADER_LOCK,
    OP_MODIFY_HIDDEN_COLUMNS,
    OP_MODIFY_ROW,
    OP_MODIFY_ROW_COLOR,
    OP_MODIFY_ROW_HEIGHT,
    OP_MODIFY_ROWS,
    OP_MODIFY_SORTS,
    OP_MODIFY_VIEW_LOCK,
    OP_MODIFY_VIEW_TYPE,
    OP_RENAME_COLUMN,
    OP_RENAME_TABLE,
    OP_RENAME_VIEW,
    OP_UPDATE_COLUMN_COLORBYS,
    OP_UPDATE_COLUMN_DESCRIPTION,
    OP_UPDATE_ROW_LINKS,
    OP_UPDATE_ROWS_LINKS,
    PARQUET_OVERWRITE_EVENTS,
    ROW_EVENTS,
    TABLE_EVENTS,
    VIEW_EVENTS,
    get_event_cat_name,
)
from .events.model import Event
from .ws_client.client import NEW_NOTIFICATION, UPDATE_DTABLE

logger = logging.getLogger(__name__)

# ACTIONS
ACTION_DELETE_PARQUET = "delete_parquet"
ACTION_OVERWRITE_PARQUET = "overwrite_parquet"


# Helper
def generate_offset():
    return str(int(time.time() * 1e3))


class Consumer:
    def __init__(
        self,
        redis_host: str = REDIS_HOST,
        redis_port: Union[int, str] = REDIS_PORT,
        key_prefix: str = KEY_PREFIX,
        seatable_url: str = SEATABLE_URL,
        seatable_username: str = SEATABLE_USERNAME,
        seatable_password: str = SEATABLE_PASSWORD,
        view_suffix: str = "__sync",
        aws_s3_access_key_id: str = AWS_S3_ACCESS_KEY_ID,
        aws_s3_secret_access_key: str = AWS_S3_SECRET_ACCESS_KEY,
        aws_s3_region_name: str = AWS_S3_REGION_NAME,
        aws_s3_bucket_name: str = AWS_S3_BUCKET_NAME,
        aws_s3_bucket_prefix: str = AWS_S3_BUCKET_PREFIX,
        batch_seconds: float = 30.0,
    ):
        # Redis to Subscribe
        self.redis_host = redis_host
        self.redis_port = int(redis_port) if redis_port else redis_port
        self.key_prefix = key_prefix

        self.redis_client = redis.Redis(host=self.redis_host, port=self.redis_port, decode_responses=True)
        self.key_update_dtable = "|".join([self.key_prefix, UPDATE_DTABLE])
        self.key_new_notification = "|".join([self.key_prefix, NEW_NOTIFICATION])

        # Seatable
        self.seatable_url = seatable_url
        self.seatable_username = seatable_username
        self.seatable_password = seatable_password
        self.view_suffix = view_suffix

        self.seatable_admin_client = AdminClient(
            seatable_url=self.seatable_url,
            seatable_username=self.seatable_username,
            seatable_password=self.seatable_password,
        )

        # AWS S3
        self.aws_s3_access_key_id = aws_s3_access_key_id
        self.aws_s3_secret_access_key = aws_s3_secret_access_key
        self.aws_s3_region_name = aws_s3_region_name
        self.aws_s3_bucket_name = aws_s3_bucket_name
        self.aws_s3_bucket_prefix = aws_s3_bucket_prefix

        self.boto_session = aioboto3.Session(
            aws_access_key_id=self.aws_s3_access_key_id,
            aws_secret_access_key=self.aws_s3_secret_access_key,
            region_name=self.aws_s3_region_name,
        )

        # batch
        self.batch_seconds = batch_seconds

        # Store
        self.watch_list = dict()
        self.do_list = dict()
        self.row_events = dict()
        self.last_offset = generate_offset()

    async def run(self):
        self.last_offset = generate_offset()
        _ = await self.init_watch_list()

        while True:
            offset = generate_offset()

            # get messages
            messages = await self.redis_client.xrange(name=self.key_update_dtable, min=self.last_offset, max=offset)

            # parse messages and submit to handlers - 순서 보장되어야 함
            for idx, msg in messages:
                base, data = orjson.loads(msg["base"]), orjson.loads(msg["data"])
                base_uuid = base["base_uuid"]

                events = event_parser(data)
                for event in events:
                    if event.op_type in TABLE_EVENTS:
                        await self.table_event_handler(idx=idx, base_uuid=base_uuid, e=event)
                    if event.op_type in VIEW_EVENTS:
                        await self.view_event_handler(idx=idx, base_uuid=base_uuid, e=event)
                    if event.op_type in COLUMN_EVENTS:
                        await self.column_event_handler(idx=idx, base_uuid=base_uuid, e=event)
                    if event.op_type in ROW_EVENTS:
                        await self.row_event_handler(idx=idx, base_uuid=base_uuid, e=event)
                    _msg = "handlers - unknown event {}".format(str(event))
                    raise KeyError(_msg)

            # sync

    ################################################################
    # Register
    ################################################################
    # init watch list
    async def init_watch_list(self):
        bases = await self.seatable_admin_client.list_bases()
        _ = await asyncio.gather(*[self.init_watch_list_for_each_base(base) for base in bases])

    # init watch list for each base
    async def init_watch_list_for_each_base(self, base: Base):
        bc = await self.seatable_admin_client.get_base_client_with_account_token_by_base_uuid(base_uuid=base.uuid)
        bigdata_status = await bc.get_bigdata_status()
        if "tables" not in bigdata_status:
            return
        for table in bigdata_status["tables"]:
            views = await bc.list_views(table["name"])
            for view in views:
                if view.type == "archive" and view.private_for is None:
                    _ = self.register_watch_list(base_uuid=base.uuid, table_id=table["id"], view=view)

    # register watch list
    def register_watch_list(self, base_uuid: str, table_id: str, view: View):
        if base_uuid not in self.watch_list:
            self.watch_list[base_uuid] = dict()
        if table_id not in self.watch_list[base_uuid]:
            self.watch_list[base_uuid][table_id] = dict()
        self.watch_list[base_uuid][table_id].update({view.id: view})

    # unregister watch list
    def unregister_watch_list(self, base_uuid: str, table_id: str, view_id: str):
        _ = self.watch_list[base_uuid][table_id].pop(view_id)
        return

    # register do list
    def register_do_list(self, base_uuid: str, table_id: str, view_id: str, action: str):
        if base_uuid not in self.do_list:
            self.do_list[base_uuid] = dict()
        if table_id not in self.do_list[base_uuid]:
            self.do_list[base_uuid][table_id] = dict()
        self.do_list[base_uuid][table_id].update({view_id: action})

    ################################################################
    # Event Handlers
    ################################################################
    # [TABLE EVENTS]
    async def table_event_handler(self, idx: str, base_uuid: str, e: Event):
        # 등록되지 않은 테이블이면 종료
        if e.table_id not in self.watch_list[base_uuid]:
            return

        # 테이블 삭제 Event
        if e.op_type in [OP_DELETE_TABLE]:
            views = [view for _, view in self.watch_list[base_uuid][e.table_id].items()]
            for view in views:
                #  (1) Watch List에서 삭제
                await self.unregister_watch_list(base_uuid=base_uuid, table_id=e.table_id, view=view.id)
                #  (2) Sink File 삭제
                await self.register_do_list(
                    base_uuid=base_uuid, table_id=e.table_id, view_id=e.view_id, action=ACTION_DELETE_PARQUET
                )
            return

        # 테이블 이름 변경 - 마지막 값만 중요
        if e.op_type in [OP_RENAME_TABLE]:
            views = [view for _, view in self.watch_list[base_uuid][e.table_id].items()]
            for view in views:
                self.register_do_list(
                    base_uuid=base_uuid, table_id=e.table_id, view_id=view.id, action=ACTION_OVERWRITE_PARQUET
                )

        # 무시
        if e.op_type in [OP_INSERT_TABLE, OP_MODIFY_HEADER_LOCK]:
            return

        # Raise Error
        _msg = "table_event_handler - miss event: {}".format(str(e))
        raise KeyError(_msg)

    # [VIEW EVENTS] Register는 이곳에서!
    async def view_event_handler(self, idx: str, base_uuid: str, e: Event):
        # REGISTER - 새로운 archive view 등록
        if e.op_type == OP_INSERT_VIEW and e.view_data.type == "archive" and e.view_data.private_for is None:
            await self.register_watch_list(base_uuid=base_uuid, table_id=e.table_id, view=e.view_data)
            return

        # 등록되지 않은 Table이면 종료
        if e.table_id not in self.watch_list[base_uuid]:
            return

        # View Type이 table -> archive 이면 등록, View Type이 archive -> table이면 등록 해제
        if e.op_type == OP_MODIFY_VIEW_TYPE:
            bc = await self.seatable_admin_client.get_base_client_with_account_token_by_base_uuid(base_uuid)
            view = await bc.get_view_by_id(table_id=e.table_id, view_id=e.view_id)
            if view.private_for:
                # private view는 pass
                return
            if e.view_type == "archive":
                _ = await self.register_watch_list(base_uuid=base_uuid, table_id=e.table_id, view=view.id)
                return
            if e.view_type == "table":
                if e.view_id in self.watch_list[base_uuid][e.table_id] and e.view_type == "table":
                    _ = await self.unregister_watch_list(base_uuid=base_uuid, table_id=e.table_id, view=view.id)
                    return
            _msg = "view_event_handler - unknown view_type '{}'.".format(e.view_type)
            raise KeyError(_msg)

        # 등록되지 않은 View면 종료
        if e.view_id not in self.watch_list[base_uuid][e.table_id]:
            return

        # View 삭제 -> 목적지에서도 삭제
        if e.op_type == OP_DELETE_VIEW:
            # (1) view 등록 해제
            _ = self.unregister_watch_list(base_uuid=base_uuid, table_id=e.table_id, view=e.view_id)
            # (2) view 삭제
            _ = self.register_do_list(
                base_uuid=base_uuid, table_id=e.table_id, view_id=e.view_id, action=ACTION_DELETE_PARQUET
            )
            return

        # View 이름 변경 -> 목적지에서 이름 변경
        if e.op_type == OP_RENAME_VIEW:
            # (1) watch list update
            self.watch_list[base_uuid][e.table_id][e.view_id].update({"name": e.view_name})
            # (2) overwrite
            _ = self.register_do_list(
                base_uuid=base_uuid, table_id=e.table_id, view_id=e.view_id, action=ACTION_OVERWRITE_PARQUET
            )

        # View Modify - Filters, Sorts, Groupbys -> 목적지 파일 덮어쓰기
        if e.op_type in [OP_MODIFY_FILTERS, OP_MODIFY_SORTS, OP_MODIFY_GROUPBYS, OP_MODIFY_HIDDEN_COLUMNS]:
            _ = self.register_do_list(
                base_uuid=base_uuid, table_id=e.table_id, view_id=e.view_id, action=ACTION_OVERWRITE_PARQUET
            )
            return

        # View Modify - Row Color, Row Heights -> Do Nothing
        if e.op_type in [OP_MODIFY_ROW_COLOR, OP_MODIFY_ROW_HEIGHT, OP_MODIFY_VIEW_LOCK]:
            # do nothing
            return

        # Raise Error
        _msg = "view_event_hander - miss event: {}".format(str(e))
        raise KeyError(_msg)

    # [COLUMN EVENT]
    async def column_event_handler(self, idx: str, base_uuid: str, e: Event):
        # 등록되어 있지 않은 Table 무시
        if e.table_id not in self.watch_list[base_uuid][e.table_id]:
            return

        if e.op_type in [OP_INSERT_COLUMN, OP_DELETE_COLUMN, OP_MODIFY_COLUMN_TYPE, OP_RENAME_COLUMN]:
            # do something
            return

        if e.op_type in [
            OP_MODIFY_COLUMN_PERMISSION,
            OP_MODIFY_COLUMN_METADATA_PERMISSION,
            OP_UPDATE_COLUMN_COLORBYS,
            OP_UPDATE_COLUMN_DESCRIPTION,
        ]:
            # do nothing
            return

        _msg = "column_event_hander - miss event: {}".format(str(e))
        raise KeyError(_msg)

    # [ROW EVENT]
    async def row_event_handler(self, idx: str, base_uuid: str, e: Event):
        # 등록되어 있지 않은 Table 무시
        if e.table_id not in self.watch_list[base_uuid][e.table_id]:
            return

        if e.op_type in [OP_INSERT_ROW, OP_DELETE_ROW, OP_APPEND_ROW, OP_MODIFY_ROW, OP_UPDATE_ROW_LINKS]:
            # Sync
            return

        if e.op_type in [OP_ARCHIVE_ROW]:
            # do nothing
            return

        _msg = "row_event_handler - miss event: {}".format(str(e))
        raise KeyError(_msg)

    ################################################################
    #
    ################################################################
    def get_obj_key(self, base_client: BaseClient, table_name: str, format: str = "parquet"):
        base_token = base_client.base_token
        keys = [self.aws_s3_bucket_prefix, base_token.group_name, base_token.base_name, table_name]
        return "/".join([*keys, ".".join(["__".join(keys), format])])

    async def overwrite_parquet_to_s3(self, base_client: BaseClient, table_name: str, parquet_version: str = "1.0"):
        offset = generate_offset()
        records = await base_client.read_table(table_name=table_name)
        tbl = pa.Table.from_pylist(records)
        with io.BytesIO() as buffer:
            pq.write_table(table=tbl, where=buffer, version=parquet_version)
            buffer.seek(0)
            content = buffer.read()
        obj_key = self.get_obj_key(base_client=base_client, table_name=table_name)
        async with self.boto_session.client("s3") as client:
            await client.upload_fileobj(io.BytesIO(content), self.aws_s3_bucket_name, obj_key)
        return offset



import asyncio
import logging
from datetime import datetime
from typing import List, Union

import aiohttp
import orjson
from pydantic import BaseModel
from tabulate import tabulate

from ..model import (
    Admin,
    ApiToken,
    Base,
    BaseToken,
    Column,
    File,
    Table,
    Team,
    User,
    Webhook,
)
from .conf import SEATABLE_URL

logger = logging.getLogger()

TABULATE_CONF = {"tablefmt": "psql", "headers": "keys"}


def parse_name(*name, delim: str = "/"):
    return [x for e in name for x in (e.split(delim) if e else [None])][: len(name)]


################################################################
# HttpClient
################################################################
class HttpClient:
    def __init__(self, seatable_url: str = SEATABLE_URL):
        self.seatable_url = seatable_url.rstrip("/")

        self.headers = {"accept": "application/json"}
        self.debug = False
        self._request = None

    async def info(self):
        async with self.session_maker() as session:
            return await self.request(session=session, method="GET", url="/server-info/")

    async def ping(self):
        async with self.session_maker() as session:
            return await self.request(session=session, method="GET", url="/api2/ping/")

    def session_maker(self, token: str = None):
        headers = self.headers.copy()
        if token:
            headers.update({"authorization": "Bearer {}".format(token)})
        return aiohttp.ClientSession(base_url=self.seatable_url, headers=headers)

    async def request(
        self,
        session: aiohttp.ClientSession,
        method: str,
        url: str,
        json: str = None,
        data: bytes = None,
        **params,
    ):
        # for debug
        self._request = {
            "method": method,
            "url": url,
            "json": json if not json else {k: v for k, v in json.items() if v},
            "data": data,
            "params": params if not params else {k: v for k, v in params.items() if v},
        }

        async with session.request(**self._request) as response:
            response.raise_for_status()
            if self.debug:
                print(response.headers)
                return await response.content()
            try:
                if response.content_type in ["application/json"]:
                    return await response.json()
                if response.content_type in ["text/html"]:
                    logger.warning(f"! content-type: {response.content_type}")
                    body = await response.text()
                    return orjson.loads(body)
                if response.content_type in [
                    "application/ms-excel",
                    "application/x-zip-compressed",
                ]:
                    content = b""
                    async for data in response.content.iter_chunked(2048):
                        content += data
                    if len(content) != response.content_length:
                        raise ValueError()
                    return File(filename=response.content_disposition.filename, content=content)

            except Exception as ex:
                raise ex

    @staticmethod
    def print(records: List[dict], tabulate_conf: dict = TABULATE_CONF):
        print(tabulate(records, **tabulate_conf))



import os

from dotenv import load_dotenv
from dasida import get_secrets

load_dotenv()

################################################################
# AWS SecretsManager
################################################################
SM_PROFILE = os.getenv("SM_PROFILE", "default")
SM_SEATABLE_ADMIN = os.getenv("SM_SEATABLE_ADMIN")
SM_AWS_S3 = os.getenv("SM_AWS_S3")

################################################################
# Redis Settings
################################################################
REDIS_HOST = os.getenv("REDIS_HOST", "localhost")
REDIS_PORT = os.getenv("REDIS_PORT", 6379)
KEY_PREFIX = os.getenv("KEY_PREFIX", "fasto")

################################################################
# SeaTable Confs
################################################################
SEATABLE_URL = os.getenv("SEATABLE_URL")
SEATABLE_USERNAME = os.getenv("SEATABLE_USERNAME")
SEATABLE_PASSWORD = os.getenv("SEATABLE_PASSWORD")

if SM_SEATABLE_ADMIN:
    secrets = get_secrets(SM_SEATABLE_ADMIN, profile_name=SM_PROFILE)
    if secrets:
        SEATABLE_URL = secrets.get("seatable_url")
        SEATABLE_USERNAME = secrets.get("seatable_username")
        SEATABLE_PASSWORD = secrets.get("seatable_password")

################################################################
# AWS S3 Confs
################################################################
AWS_S3_BUCKET_NAME = os.getenv("AWS_S3_BUCKET_NAME")
AWS_S3_BUCKET_PREFIX = os.getenv("AWS_S3_BUCKET_PREFIX", "fasto")

AWS_S3_ACCESS_KEY_ID = os.getenv("AWS_S3_ACCESS_KEY_ID")
AWS_S3_SECRET_ACCESS_KEY = os.getenv("AWS_S3_SECRET_ACCESS_KEY")
AWS_S3_REGION_NAME = os.getenv("AWS_S3_REGION_NAME", "ap-northeast-2")

if SM_AWS_S3:
    secrets = get_secrets(SM_AWS_S3, profile_name=SM_PROFILE)
    if secrets:
        AWS_S3_ACCESS_KEY_ID = secrets.get("aws_access_key_id")
        AWS_S3_SECRET_ACCESS_KEY = secrets.get("aws_secret_access_key")



import asyncio
import logging
from datetime import datetime
from typing import List, Union

import aiohttp
import orjson
import requests
from pydantic import BaseModel
from tabulate import tabulate

from ..model import (
    DTABLE_ICON_COLORS,
    DTABLE_ICON_LIST,
    Admin,
    ApiToken,
    Base,
    BaseToken,
    Column,
    Table,
    Team,
    User,
    Webhook,
)
from .base import BaseClient
from .conf import SEATABLE_ACCOUNT_TOKEN, SEATABLE_PASSWORD, SEATABLE_URL, SEATABLE_USERNAME
from .core import TABULATE_CONF, HttpClient

logger = logging.getLogger()


################################################################
# AccountClient
################################################################
class AccountClient(HttpClient):
    def __init__(
        self,
        seatable_url: str = SEATABLE_URL,
        seatable_username: str = SEATABLE_USERNAME,
        seatable_password: str = SEATABLE_PASSWORD,
    ):
        super().__init__(seatable_url=seatable_url)
        self.username = seatable_username
        self.password = seatable_password
        self.account_token = None

        self.is_admin = False

        # do login
        self.login()

    def login(self):
        auth_url = self.seatable_url + "/api2/auth-token/"
        response = requests.post(
            auth_url, json={"username": self.username, "password": self.password}
        )
        response.raise_for_status()
        results = response.json()
        self.account_token = results["token"]

    ################################################################
    # AUTHENTICATION - API TOKEN
    ################################################################
    # [API TOKEN] list api tokens
    async def list_api_tokens(self, workspace_id: str, base_name: str, model: BaseModel = ApiToken):
        """
        [NOTE]
         workspace id : group = 1 : 1
        """
        METHOD = "GET"
        URL = f"/api/v2.1/workspace/{workspace_id}/dtable/{base_name}/api-tokens/"
        ITEM = "api_tokens"

        async with self.session_maker(token=self.account_token) as session:
            response = await self.request(session=session, method=METHOD, url=URL)
            results = response[ITEM]

        if model:
            results = [model(**x) for x in results]

        return results

    # [API TOKEN] create api token
    async def get_or_create_api_token(
        self,
        workspace_id: str,
        base_name: str,
        app_name: str,
        permission: str = "rw",
        model: BaseModel = ApiToken,
    ):
        """
        [NOTE]
         "bad request" returns if app_name is already exists.
        """
        api_tokens = await self.list_api_tokens(workspace_id=workspace_id, base_name=base_name)
        for api_token in api_tokens:
            if api_token.app_name == app_name:
                return api_token

        METHOD = "POST"
        URL = f"/api/v2.1/workspace/{workspace_id}/dtable/{base_name}/api-tokens/"
        JSON = {"app_name": app_name, "permission": permission}

        async with self.session_maker(token=self.account_token) as session:
            response = await self.request(session=session, method=METHOD, url=URL, json=JSON)
            results = model(**response) if model else response

        return results

    # [API TOKEN] create temporary api token
    async def create_temp_api_token(
        self, workspace_id: str, base_name: str, model: BaseModel = ApiToken
    ):
        METHOD = "GET"
        URL = f"/api/v2.1/workspace/{workspace_id}/dtable/{base_name}/temp-api-token/"
        ITEM = "api_token"

        async with self.session_maker(token=self.account_token) as session:
            response = await self.request(session=session, method=METHOD, url=URL)
            results = response[ITEM]

        if model:
            now = datetime.now()
            results = model(
                app_name="__temp_token",
                api_token=results,
                generated_by="__temp_token",
                generated_at=now,
                last_access=now,
                permission="r",
            )

        return results

    # [API TOKEN] update api token
    async def update_api_token(
        self,
        workspace_id: str,
        base_name: str,
        app_name: str,
        permission: str = "rw",
        model: BaseModel = ApiToken,
    ):
        METHOD = "PUT"
        URL = f"/api/v2.1/workspace/{workspace_id}/dtable/{base_name}/api-tokens/{app_name}"
        JSON = {"permission": permission}

        async with self.session_maker(token=self.account_token) as session:
            response = await self.request(session=session, method=METHOD, url=URL, json=JSON)
            results = model(**response) if model else response

        return results

    # [API TOKEN] delete api token
    async def delete_api_token(self, workspace_id: str, base_name: str, app_name: str):
        METHOD = "DELETE"
        URL = f"/api/v2.1/workspace/{workspace_id}/dtable/{base_name}/api-tokens/{app_name}"
        ITEM = "success"

        async with self.session_maker(token=self.account_token) as session:
            response = await self.request(session=session, method=METHOD, url=URL)
            results = response[ITEM]

        return results

    ################################################################
    # AUTHENTICATION - BASE TOKEN
    ################################################################
    # [BASE TOKEN] get base token with account token
    async def get_base_token_with_account_token(
        self,
        workspace_id: str = None,
        base_name: str = None,
        model: BaseModel = BaseToken,
    ):
        METHOD = "GET"
        URL = f"/api/v2.1/workspace/{workspace_id}/dtable/{base_name}/access-token/"

        async with self.session_maker(token=self.account_token) as session:
            results = await self.request(session=session, method=METHOD, url=URL)
        if model:
            results = model(**results)

        return results

    # [BASE TOKEN] get base token with api token
    async def get_base_token_with_api_token(self, api_token: str, model: BaseModel = BaseToken):
        METHOD = "GET"
        URL = "/api/v2.1/dtable/app-access-token/"

        async with self.session_maker(token=api_token) as session:
            results = await self.request(session=session, method=METHOD, url=URL)
        if model:
            results = model(**results)

        return results

    # [BASE TOKEN] get base token with invite link
    async def get_base_token_with_invite_link(self, link: str, model: BaseModel = BaseToken):
        link = link.rsplit("/links/", 1)[-1].strip("/")
        METHOD = "GET"
        URL = "/api/v2.1/dtable/share-link-access-token/"

        async with self.session_maker(token=link) as session:
            results = await self.request(session=session, method=METHOD, url=URL)
        if model:
            results = model(**results)

        return results

    # [BASE TOKEN] get base token with external link
    async def get_base_token_with_external_link(self, link: str, model: BaseModel = BaseToken):
        link = link.rsplit("/external-links/", 1)[-1].strip("/")
        METHOD = "GET"
        URL = f"/api/v2.1/external-link-tokens/{link}/access-token/"

        async with self.session_maker(token=link) as session:
            results = await self.request(session=session, method=METHOD, url=URL)
        if model:
            results = model(**results)

        return results

    ################################################################
    # (CUSTOM) GET BASE CLIENT
    ################################################################
    # [BASE CLIENT] (custom) get base client with account token
    async def get_base_client_with_account_token(self, workspace_id: str, base_name: str):
        base_token = await self.get_base_token_with_account_token(
            workspace_id=workspace_id, base_name=base_name
        )
        return BaseClient(base_token=base_token)



from typing import Any, List, Union

from pydantic import BaseModel, Extra, Field


# EVENT
class Event(BaseModel):
    class Config:
        extra = Extra.forbid


# OPTION
class Option(BaseModel):
    class Config:
        extra = Extra.forbid


################################################################
# Column
################################################################
class Filter(Option):
    column_key: str
    filter_predicate: str
    filter_term: str


class Sort(Option):
    column_key: str
    sort_type: str


class Groupby(Option):
    column_key: str
    sort_type: str
    count_type: str


################################################################
# Column
################################################################
class Column(Option):
    rowType: str = None
    key: str  # "4xVF",
    type: str  # "text",
    name: str  # "hello",
    editable: bool  # True,
    width: int  # 200,
    resizable: bool  # True,
    draggable: bool  # True,
    data: dict = None
    permission_type: str = None  # ""
    permitted_users: List[str] = None  # []
    edit_metadata_permission_type: str  # ""
    edit_metadata_permitted_users: list  # []
    description: str = None
    editor: dict = None
    formatter: dict = None
    idx: int = None
    left: int = None
    last_frozen: bool = None


class ColumnPermission(Option):
    permission_type: str
    permitted_users: List[str]


class ColumnMetadataPermission(Option):
    edit_metadata_permission_type: str
    edit_metadata_permitted_users: List[str]


################################################################
# View
################################################################
class View(Option):
    id: str = Field(alias="_id")
    name: str
    type: str
    private_for: str = None
    is_locked: bool
    row_height: str
    filter_conjunction: str
    filters: List[dict]
    sorts: List[dict]
    groupbys: List[dict]
    colorbys: dict
    hidden_columns: list
    rows: list
    formula_rows: dict
    link_rows: dict
    summaries: dict
    colors: dict
    column_colors: dict
    groups: list


################################################################
# Table
################################################################
class Table(Option):
    id: str = Field(alias="_id")
    name: str
    is_header_locked: bool
    header_settings: dict
    summary_configs: dict
    columns: List[Column] = None
    rows: List[dict]
    view_structure: dict
    views: List[View] = None
    id_row_map: dict



import logging
from datetime import datetime, date
from typing import Any, List, Union


from ..model import Table
from .const import DT_FMT, SYSTEM_FIELDS, TZ

logger = logging.getLogger(__name__)


################################################################
# Converter
################################################################
class ToPythonDict:
    def __init__(self, table: Table, users: dict = None):
        self.table = table
        self.users = users

        self.columns = {
            **{column.name: {"column_type": column.type, "column_data": column.data} for column in table.columns},
            **SYSTEM_FIELDS,
        }

        self.user_map = (
            {user.email: f"{user.name} ({user.contact_email})" for user in self.users} if self.users else None
        )
        self.row_id_map = {column.key: column.name for column in table.columns}

    def __call__(self, row):
        if row is None:
            return
        return {
            column: getattr(self, self.columns[column]["column_type"].replace("-", "_"))(
                value=row[column], data=self.columns[column].get("column_data")
            )
            for column in self.columns
            if column in row
        }

    def checkbox(self, value, data: dict = None) -> bool:
        return value

    def text(self, value: str, data: dict = None) -> str:
        return value

    def button(self, value: str, data: dict = None) -> str:
        return value

    def long_text(self, value: str, data: dict = None) -> str:
        return value

    def email(self, value: str, data: dict = None) -> str:
        return value

    def url(self, value: str, data: dict = None) -> str:
        return value

    def rate(self, value: int, data: dict = None) -> int:
        return value

    def number(self, value: Union[int, float], data: dict = None) -> Union[int, float]:
        if not value:
            return value
        if data and data.get("enable_precision") and data["precision"] == 0:
            return int(value)
        return float(value)

    def date(self, value: str, data: dict = None) -> Union[date, datetime]:
        if not value:
            return value
        if data and data["format"] == "YYYY-MM-DD":
            return date.fromisoformat(value[:10])
        if value.endswith("Z"):
            value = value.replace("Z", "+00:00", 1)
        try:
            dt = datetime.strptime(value, DT_FMT)
        except ValueError as ex:
            dt = datetime.fromisoformat(value)
        return dt.astimezone(TZ)

    def duration(self, value: str, data: dict = None) -> int:
        """
        return seconds
        """
        return value

    def ctime(self, value, data: dict = None):
        return self.date(value, data)

    def mtime(self, value, data: dict = None):
        return self.date(value, data)

    def single_select(self, value: str, data: dict = None) -> str:
        return value

    def multiple_select(self, value: List[str], data: dict = None) -> List[str]:
        return value

    def link(self, value: List[Any], data: dict = None) -> list:
        if not value:
            return value
        value = [x["display_value"] for x in value]
        if not value:
            return
        if data:
            if "array_type" in data and data["array_type"] == "single-select":
                kv = {x["id"]: x["name"] for x in data["array_data"]["options"]}
                value = [kv[x] if x in kv else x for x in value]
            if "is_multiple" in data and not data["is_multiple"]:
                value = value[0]
        return value

    def link_formula(self, value, data: dict = None):
        if not value:
            return
        if data:
            if "array_type" in data:
                array_data = data.get("array_data")
                if not isinstance(value, list):
                    value = [value]
                value = [getattr(self, data["array_type"])(x, array_data) for x in value]
            link_column = self.columns[self.row_id_map[data["link_column_key"]]]
            if not link_column["column_data"]["is_multiple"]:
                value = value[0]
        return value

    def user(self, user: str):
        if not self.user_map:
            return user
        if user in self.user_map:
            return self.user_map[user]
        return user

    def collaborator(self, value: List[str], data: dict = None) -> List[str]:
        if not value:
            return value
        return [self.user(x) for x in value]

    def creator(self, value: str, data: dict = None) -> str:
        return self.user(value)

    def last_modifier(self, value: str, data: dict = None) -> str:
        return self.user(value)

    def file(self, value, data: dict = None):
        if value is None:
            return
        return [x["url"] for x in value]

    def image(self, value, data: dict = None):
        return value

    def formula(self, value, data: dict = None):
        if data:
            try:
                value = getattr(self, "{}".format(data["result_type"]))(value)
            except Exception as ex:
                if value != "#VALUE!":
                    raise ex
                value = None
        return value

    def auto_number(self, value, data: dict = None):
        return value



import io
from typing import List

import aioboto3
import pyarrow as pa
import pyarrow.parquet as pq

from plantable.client.base import BaseClient
from .conf import AWS_S3_BUCKET_NAME, AWS_S3_BUCKET_PREFIX


# Generate Filename
def generate_filename(
    workspace_name: str, base_name: str, table_name: str, view_name: str = None, format: str = "parquet"
):
    names = [workspace_name, base_name, table_name]
    if view_name:
        names.append(view_name)
    return ".".join(["_".join(names), format])


# Generate S3 Object Key
def generate_obj_key(
    format: str,
    workspace_name: str,
    base_name: str,
    table_name: str,
    view_name: str,
    group: str = None,
    aws_s3_bucket_prefix: str = AWS_S3_BUCKET_PREFIX,
) -> str:
    keys = [
        aws_s3_bucket_prefix,
        format,
        workspace_name,
        base_name,
        table_name,
    ]

    # append prefix
    if group:
        keys.append(f"group={group}")
    keys.append(f"view={view_name}")

    # append filename
    names = [workspace_name, base_name, table_name]
    if view_name:
        names.append(view_name)
    filename = ".".join(["_".join(names), format])
    keys.append(filename)

    return "/".join(keys)


# Python List to Parquet Bytes
def pylist_to_parquet(records: List[dict], version: str = "1.0") -> bytes:
    tbl = pa.Table.from_pylist(records)
    with io.BytesIO() as buffer:
        pq.write_table(table=tbl, where=buffer, version=version)
        buffer.seek(0)
        content = buffer.read()
    return content


# Read Table using BaseClient as Parquet
async def table_to_parquet(client: BaseClient, table_name: str, modified_before: str, modified_after: str):
    records = await client.read_table(
        table_name=table_name,
        modified_before=modified_before,
        modified_after=modified_after,
    )
    return pylist_to_parquet(records)


# Read Table using BaseClient as Parquet
async def view_to_parquet(client: BaseClient, table_name: str, view_name: str):
    records = await client.read_view(table_name=table_name, view_name=view_name)
    return pylist_to_parquet(records)


# Upload to S3
async def upload_view_to_s3(
    session: aioboto3.Session,
    content: bytes,
    format: str,
    prod: bool,
    workspace_name: str,
    base_name: str,
    table_name: str,
    view_name: str = None,
    group: str = None,
):
    obj_key = generate_obj_key(
        format=format,
        prod=prod,
        workspace_name=workspace_name,
        base_name=base_name,
        table_name=table_name,
        view_name=view_name,
        group=group,
    )

    async with session.client("s3") as client:
        await client.upload_fileobj(io.BytesIO(content), AWS_S3_BUCKET_NAME, obj_key)

    return {"bucket": AWS_S3_BUCKET_NAME, "key": obj_key, "size": len(content)}


# Download from S3
async def download_from_s3(
    session: aioboto3.Session,
    bucket: str,
    key: str,
):
    async with session.resource("s3") as client:
        object = await client.Object(bucket, key)
        results = await object.get()
        content = await results["Body"].read()

    return content



import asyncio
import logging
import time
from datetime import datetime, timedelta
from typing import Callable, Union

import engineio
import orjson
import six
import socketio
from socketio.exceptions import ConnectionError

from plantable.client import AdminClient

from ..conf import SEATABLE_PASSWORD, SEATABLE_URL, SEATABLE_USERNAME

logger = logging.getLogger(__file__)

JOIN_ROOM = "join-room"
CONNECT = "connect"
DISCONNECT = "disconnect"
IO_DISCONNECT = "io-disconnect"
CONNECT_ERROR = "connect_error"
UPDATE_DTABLE = "update-dtable"
NEW_NOTIFICATION = "new-notification"


################################################################
# Websocket
################################################################
class BaseWebsocketClient(socketio.AsyncClient):
    """
    [NOTE] admin user만 사용할 수 있도록 작성되어 있음
     - AdminClient 사용하여 BaseToken 얻는 구조
    """

    def __init__(
        self,
        group_name_or_id: Union[int, str],
        base_name: str,
        seatable_url: str = SEATABLE_URL,
        seatable_username: str = SEATABLE_USERNAME,
        seatable_password: str = SEATABLE_PASSWORD,
        handler: Callable = None,
        request_timeout: int = 30,
        reconnection_attempts=40,
        reconnection_delay=3,
    ):
        # account
        self.seatable_url = seatable_url
        self.seatable_username = seatable_username
        self.seatable_password = seatable_password

        # group & base
        self.group_name_or_id = group_name_or_id
        self.base_name = base_name

        # handler
        self.handler = handler

        # websocket io
        super().__init__(
            request_timeout=request_timeout,
            reconnection_attempts=reconnection_attempts,
            reconnection_delay=reconnection_delay,
            reconnection_delay_max=reconnection_delay,
            randomization_factor=0,
        )

        # plantable client
        self.admin_client = AdminClient(
            seatable_url=self.seatable_url,
            seatable_username=self.seatable_username,
            seatable_password=self.seatable_password,
        )

        # base token
        self.base_token = None
        self.websocket_url = None
        self.key = None

    async def update_base_token(self):
        await self.admin_client.ensure_group_member(group_name_or_id=self.group_name_or_id)
        self.base_token = await self.admin_client.get_base_token_with_account_token(
            group_name_or_id=self.group_name_or_id, base_name=self.base_name
        )
        self.websocket_url = self.seatable_url + f"?dtable_uuid={self.base_token.dtable_uuid}"
        self.key = {
            "workspace_id": self.base_token.workspace_id,
            "group_id": self.base_token.group_id,
            "group_name": self.base_token.group_name,
            "base_uuid": self.base_token.dtable_uuid,
            "base_name": self.base_token.base_name,
        }

    async def run(self, on_update: Callable = None, on_notification: Callable = None):
        try:
            self.on(CONNECT, self.on_connect)
            self.on(DISCONNECT, self.on_disconnect)
            self.on(IO_DISCONNECT, self.on_io_disconnect)
            self.on(CONNECT_ERROR, self.on_connect_error)
            self.on(UPDATE_DTABLE, on_update or self.on_update)
            self.on(NEW_NOTIFICATION, on_notification or self.on_notification)
            await self.update_base_token()
            await self.connect(url=self.websocket_url)
            await self.wait()
        except asyncio.CancelledError as ex:
            await self.disconnect()
            raise ex

    async def on_connect(self):
        if not self.base_token or datetime.now() >= self.base_token.generated_at + timedelta(days=3):
            self.update_base_token()
        await self.emit(JOIN_ROOM, (self.base_token.dtable_uuid, self.base_token.access_token))
        _msg = f"[ SeaTable SocketIO connection established ] {self.key['group_name']}/{self.key['base_name']}"
        logger.info(_msg)

    async def on_disconnect(self):
        logger.info("[ SeaTable SocketIO connection dropped ]")

    async def on_io_disconnect(self, sleep=3):
        logger.warning("[ SeaTable SocketIO connection disconnected ]")
        time.sleep(sleep)
        self.update_base_token()
        self.connect(self.websocket_url)

    async def on_connect_error(self, error_msg):
        _msg = f"[ SeaTable SocketIO connection error ] {error_msg}"
        logger.error(_msg)

    async def on_update(self, data, index, *args):
        msg = {"index": index, "base": orjson.dumps(self.key), "data": data}
        if self.handler:
            await self.handler(UPDATE_DTABLE, msg)
        else:
            print(UPDATE_DTABLE, msg)

    async def on_notification(self, data, index, *args):
        msg = {"index": index, "base": orjson.dumps(self.key), "data": data}
        if self.handler:
            await self.handler(NEW_NOTIFICATION, msg)
        else:
            print(NEW_NOTIFICATION, msg)

    ################################################################
    # Override Some Methods
    ################################################################
    async def _handle_disconnect(self, namespace):
        """io server disconnect"""
        self.logger.info("Engine.IO connection disconnected")
        if not self.connected:
            return
        self.disconnect()
        namespace = namespace or "/"
        self._trigger_event("io-disconnect", namespace=namespace)

    async def connect(self, url, headers={}, transports=None, namespaces=None, socketio_path="socket.io"):
        self.connection_url = url
        self.connection_headers = headers
        self.connection_transports = transports
        self.connection_namespaces = namespaces
        self.socketio_path = socketio_path

        if namespaces is None:
            namespaces = set(self.handlers.keys()).union(set(self.namespace_handlers.keys()))
        elif isinstance(namespaces, six.string_types):
            namespaces = [namespaces]
            self.connection_namespaces = namespaces
        self.namespaces = [n for n in namespaces if n != "/"]
        try:
            await self.eio.connect(url, headers=headers, transports=transports, engineio_path=socketio_path)
        except engineio.exceptions.ConnectionError as exc:
            _msg = f"{self.key['group_name']}/{self.key['base_name']} - {exc.args[0]}"
            await self._trigger_event("connect_error", "/", _msg)
            raise ConnectionError(exc.args[0])
        self.connected = True



from datetime import datetime
from typing import List

from pydantic import BaseModel, Field, validator

from .core import _Model

__all__ = ["AccountInfo", "Admin", "Team", "ApiToken", "BaseToken", "Webhook"]


class AccountInfo(_Model):
    space_usage: str  # '0%'
    total: int  # -2
    usage: int  # 0
    row_usage_rate: str  # '0%'
    row_total: int  # -1
    row_usage: int  # 0
    avatar_url: str  # 'https://seatable.jongno.life/media/avatars/default.png'
    email: str  # '2926d3fa3a364558bac8a550811dbe0e@auth.local'
    name: str  # 'admin'
    login_id: str = None  # ''
    contact_email: str  # 'woojin.cho@gmail.com'
    institution: str = None  # ''
    is_staff: bool  # True
    enable_subscription: bool  # False
    dtable_updates_email_interval: int  # 0
    dtable_collaborate_email_interval: int  # 0


class Admin(_Model):
    email: str  # '2926d3fa3a364558bac8a550811dbe0e@auth.local'
    name: str  # 'admin'
    contact_email: str  # 'woojin.cho@gmail.com'
    login_id: str = None  # ''
    is_staff: bool  # True
    is_active: bool  # True
    storage_usage: int  # 0
    rows_count: int  # 0
    create_time: datetime  # '2023-05-21T03:04:26+00:00'
    last_login: datetime  # '2023-05-28T11:42:01+00:00'
    admin_role: str  # 'default_admin'


class Team(_Model):
    org_id: str  # 1
    org_name: str  # "Test-Admin"
    ctime: datetime  # "2020-06-01T12:46:26+00:00"
    org_url_prefix: str  # "org_8hz6uh"
    role: str  # "org_default"
    creator_email: str  # "8ca1997823b44dffbaa51e0dd0c35ac0@auth.local"
    creator_name: str  # "Christoph Dyllick"
    creator_contact_email: str  # "christoph@example.com"
    quota: int  # -2
    storage_usage: int  # 0
    storage_quota: int  # 1000000000
    max_user_number: int  # 25
    rows_count: int  # 0
    row_limit: int  # 200


class ApiToken(_Model):
    app_name: str  # 'n8n', null when created with account token
    api_token: str  # 'f5ca15bf3bb64101be0a03a57feaf2289f494701'
    generated_by: str  # '2926d3fa3a364558bac8a550811dbe0e@auth.local'
    generated_at: datetime  # '2023-05-28T13:00:31+00:00'
    last_access: datetime  # '2023-05-28T13:15:10+00:00'
    permission: str  # 'rw


class BaseToken(_Model):
    app_name: str = None  # 'test-api'
    access_token: str  # 'eyJ0eXAiOiJKV1QiLCJhbGciOiJIUzI1NiJ9.eyJleHAiOjE2ODU2MDg3ODcsImR0YWJsZV91dWlkIjoiMTY2NDI0YWQtYjAyMy00N2EwLTlhMzUtNzYwNzdmNWI2MjliIiwidXNlcm5hbWUiOiIiLCJwZXJtaXNzaW9uIjoicnciLCJhcHBfbmFtZSI6InRlc3QtYXBpIn0.djJ7NW67UicDmo35UodS5UJBQWydqAvt-euo1TzM2rY'
    dtable_uuid: str  # '166424ad-b023-47a0-9a35-76077f5b629b'
    dtable_server: str  # 'https://seatable.jongno.life/dtable-server/'
    dtable_socket: str  # 'https://seatable.jongno.life/'
    dtable_db: str = None  # 'https://seatable.jongno.life/dtable-db/'
    workspace_id: int = None  # 3
    group_id: int = None  # (manually added)
    group_name: str = None  # (manually added)
    base_name: str = Field(None, alias="dtable_name")  # 'employee
    generated_at: datetime = datetime.now()  # (manually added)


class Webhook(_Model):
    id: int
    dtable_uuid: str
    url: str
    creator: str
    created_at: datetime
    is_valid: bool
    settings: dict = None



import asyncio
import logging
import time
from datetime import datetime
from typing import Callable, List, Union

import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq
import requests
import socketio
from fastapi import HTTPException, status
from pydantic import BaseModel
from pypika import MySQLQuery as PikaQuery
from pypika import Table as PikaTable
from pypika.dialects import QueryBuilder
from tabulate import tabulate

from ..model import (
    DTABLE_ICON_COLORS,
    DTABLE_ICON_LIST,
    Admin,
    ApiToken,
    Base,
    BaseActivity,
    BaseToken,
    Column,
    Metadata,
    SelectOption,
    Table,
    Team,
    User,
    UserInfo,
    View,
    Webhook,
)
from ..serde import ToPythonDict
from ..column import SeaTableType
from .conf import SEATABLE_URL
from .core import TABULATE_CONF, HttpClient
from .exception import MoreRows

logger = logging.getLogger()


################################################################
# Helpers
################################################################
# divide chunks
def divide_chunks(x: list, chunk_size: int):
    for i in range(0, len(x), chunk_size):
        yield x[i : i + chunk_size]


################################################################
# BaseClient
################################################################
class BaseClient(HttpClient):
    def __init__(
        self,
        seatable_url: str = SEATABLE_URL,
        api_token: str = None,
        base_token: BaseToken = None,
    ):
        super().__init__(seatable_url=seatable_url.rstrip("/"))

        self.base_token = base_token
        self.api_token = api_token

        if api_token:
            auth_url = self.seatable_url + "/api/v2.1/dtable/app-access-token/"
            response = requests.get(auth_url, headers={"Authorization": f"Token {self.api_token}"})
            try:
                response.raise_for_status()
            except requests.exceptions.HTTPError as ex:
                error_msg = response.json()["error_msg"]
                if error_msg in ["Permission denied."]:
                    raise HTTPException(
                        status_code=status.HTTP_401_UNAUTHORIZED,
                        detail="Wrong base token!",
                    )
                raise ex
            results = response.json()
            self.base_token = BaseToken(**results)

        self.dtable_uuid = self.base_token.dtable_uuid
        self.workspace_id = self.base_token.workspace_id
        self.group_id = self.base_token.group_id
        self.group_name = self.base_token.group_name
        self.base_name = self.base_token.base_name

    ################################################################
    # BASE INFO
    ################################################################
    # Get Base Info
    async def get_base_info(self):
        METHOD = "GET"
        URL = f"/dtable-server/dtables/{self.base_token.dtable_uuid}"

        async with self.session_maker(token=self.base_token.access_token) as session:
            results = await self.request(session=session, method=METHOD, url=URL)

        return results

    # Get Metadata
    async def get_metadata(self, model: BaseModel = Metadata):
        METHOD = "GET"
        URL = f"/dtable-server/api/v1/dtables/{self.base_token.dtable_uuid}/metadata/"
        ITEM = "metadata"

        async with self.session_maker(token=self.base_token.access_token) as session:
            response = await self.request(session=session, method=METHOD, url=URL)
            results = response[ITEM]

        if model:
            results = model(**results)

        return results

    # (CUSTOM) List Tables
    async def list_tables(self):
        metadata = await self.get_metadata()
        tables = metadata.tables

        return tables

    # (CUSTOM) Get Table
    async def get_table(self, table_name: str):
        tables = await self.list_tables()
        for table in tables:
            if table.name == table_name:
                return table
        else:
            raise KeyError()

    # (CUSTOM) Get Table by ID
    async def get_table_by_id(self, table_id: str):
        tables = await self.list_tables()
        for table in tables:
            if table.id == table_id:
                return table
        else:
            raise KeyError()

    # (CUSTOM) Get Names by ID
    async def get_names_by_ids(self, table_id: str, view_id: str):
        table = await self.get_table_by_id(table_id=table_id)
        views = await self.list_views(table_name=table.name)
        for view in views:
            if view.id == view_id:
                break
        else:
            raise KeyError

        return table.name, view.name

    # Get Big Data Status
    async def get_bigdata_status(self):
        METHOD = "GET"
        URL = f"/dtable-db/api/v1/base-info/{self.base_token.dtable_uuid}/"

        async with self.session_maker(token=self.base_token.access_token) as session:
            results = await self.request(session=session, method=METHOD, url=URL)

        return results

    # List Collaborators
    async def list_collaborators(self, model: BaseModel = UserInfo):
        METHOD = "GET"
        URL = f"/dtable-server/api/v1/dtables/{self.base_token.dtable_uuid}/related-users/"
        ITEM = "user_list"

        async with self.session_maker(token=self.base_token.access_token) as session:
            response = await self.request(session=session, method=METHOD, url=URL)
            results = response[ITEM]

        if model:
            results = [model(**x) for x in results]

        return results

    # (CUSTOM) ls
    async def ls(self, table_name: str = None):
        metadata = await self.get_metadata()
        tables = metadata.tables
        if table_name:
            for table in tables:
                if table.name == table_name:
                    break
            else:
                raise KeyError()
            columns = [{"key": c.key, "name": c.name, "type": c.type} for c in table.columns]
            print(tabulate(columns, **TABULATE_CONF))
            return
        _tables = list()
        for table in tables:
            _n = len(table.columns)
            _columns = ", ".join(c.name for c in table.columns)
            if len(_columns) > 50:
                _columns = _columns[:50] + "..."
            _columns += f" ({_n})"
            _tables += [
                {
                    "id": table.id,
                    "name": table.name,
                    "views": ", ".join([v.name for v in table.views]),
                    "columns": _columns,
                },
            ]
        print(tabulate(_tables, **TABULATE_CONF))

    ################################################################
    # ROWS
    ################################################################
    # List Rows (Table, with SQL)
    async def list_rows_with_sql(self, sql: Union[str, QueryBuilder], convert_keys: bool = True):
        """
        [NOTE]
         default LIMIT 100 when not LIMIT is given!
         max LIMIT 10000!
        """
        METHOD = "POST"
        URL = f"/dtable-db/api/v1/query/{self.base_token.dtable_uuid}/"

        json = {
            "sql": sql.get_sql() if isinstance(sql, QueryBuilder) else sql,
            "convert_keys": convert_keys,
        }
        SUCCESS = "success"
        ITEM = "results"

        async with self.session_maker(token=self.base_token.access_token) as session:
            response = await self.request(session=session, method=METHOD, url=URL, json=json)
            if not response[SUCCESS]:
                raise Exception(response)
            results = response[ITEM]

        return results

    # List Rows (View)
    # [NOTE] 4.1에서 첫 Row에 없는 값은 안 읽어오는 이슈
    async def list_rows(
        self,
        table_name: str,
        view_name: str,
        convert_link_id: bool = False,
        order_by: str = None,
        direction: str = "asc",
        start: int = 0,
        limit: int = None,
    ):
        MAX_LIMIT = 1000

        METHOD = "GET"
        URL = f"/dtable-server/api/v1/dtables/{self.base_token.dtable_uuid}/rows/"
        ITEM = "rows"

        get_all_rows = False
        if not limit:
            get_all_rows = True
            limit = limit or MAX_LIMIT

        params = {
            "table_name": table_name,
            "view_name": view_name,
            "convert_link_id": str(convert_link_id).lower(),
            "order_by": order_by,
            "direction": direction,
            "start": start,
            "limit": limit,
        }

        async with self.session_maker(token=self.base_token.access_token) as session:
            response = await self.request(session=session, method=METHOD, url=URL, **params)
            response = response[ITEM]
            results = response

            # pagination
            if get_all_rows:
                while len(response) == limit:
                    params.update({"start": params["start"] + limit})
                    response = await self.request(session=session, method=METHOD, url=URL, **params)
                    response = response[ITEM]
                    results += response

        return results

    # (CUSTOM) List Rows by ID
    async def list_rows_by_id(
        self,
        table_id: str,
        view_id: str,
        convert_link_id: bool = False,
        order_by: str = None,
        direction: str = "asc",
        start: int = 0,
        limit: int = None,
    ):
        table_name, view_name = self.get_names_by_ids(table_id=table_id, view_id=view_id)
        return await self.list_rows(
            table_name=table_name,
            view_name=view_name,
            convert_link_id=convert_link_id,
            order_by=order_by,
            direction=direction,
            start=start,
            limit=limit,
        )

    # Add Row
    async def add_row(
        self,
        table_name: str,
        row: dict,
        anchor_row_id: str = None,
        row_insert_position: str = "insert_below",
    ):
        # insert_below or insert_above
        METHOD = "POST"
        URL = f"/dtable-server/api/v1/dtables/{self.base_token.dtable_uuid}/rows/"

        json = {"table_name": table_name, "row": row}
        if anchor_row_id:
            json.update(
                {
                    "ahchor_row_id": anchor_row_id,
                    "row_insert_position": row_insert_position,
                }
            )

        # add select options if not exists
        _ = await self.add_select_options_if_not_exists(table_name=table_name, rows=[row])

        async with self.session_maker(token=self.base_token.access_token) as session:
            results = await self.request(session=session, method=METHOD, url=URL, json=json)

        return results

    # Update Row
    async def update_row(self, table_name: str, row_id: str, row: dict):
        # NOT WORKING
        METHOD = "PUT"
        URL = f"/dtable-server/api/v1/dtables/{self.base_token.dtable_uuid}/rows/"
        ITEM = "success"

        json = {"table_name": table_name, "row_id": row_id, "row": row}

        # add select options if not exists
        _ = await self.add_select_options_if_not_exists(table_name=table_name, rows=[row])

        async with self.session_maker(token=self.base_token.access_token) as session:
            response = await self.request(session=session, method=METHOD, url=URL, json=json)
            results = response[ITEM]

        return results

    # Delete Row
    async def delete_row(self, table_name: str, row_id: str):
        METHOD = "DELETE"
        URL = f"/dtable-server/api/v1/dtables/{self.base_token.dtable_uuid}/rows/"

        json = {"table_name": table_name, "row_id": row_id}

        async with self.session_maker(token=self.base_token.access_token) as session:
            results = await self.request(session=session, method=METHOD, url=URL, json=json)

        return results

    # Get Row
    async def get_row(self, table_name: str, row_id: str, convert: bool = False):
        # NOT WORKING
        METHOD = "GET"
        URL = f"/dtable-server/api/v1/dtables/{self.base_token.dtable_uuid}/rows/{row_id}/"

        params = {"table_name": table_name, "convert": str(convert).lower()}

        async with self.session_maker(token=self.base_token.access_token) as session:
            results = await self.request(session=session, method=METHOD, url=URL, **params)

        return results

    # Append Rows
    async def append_rows(self, table_name: str, rows: List[dict]):
        # insert_below or insert_above
        METHOD = "POST"
        URL = f"/dtable-server/api/v1/dtables/{self.base_token.dtable_uuid}/batch-append-rows/"

        # add select options if not exists
        _ = await self.add_select_options_if_not_exists(table_name=table_name, rows=rows)

        # divide chunk - [NOTE] 1000 rows까지만 됨
        UPDATE_LIMIT = 1000
        chunks = divide_chunks(rows, UPDATE_LIMIT)
        list_json = [{"table_name": table_name, "rows": chunk} for chunk in chunks]

        async with self.session_maker(token=self.base_token.access_token) as session:
            coros = [self.request(session=session, method=METHOD, url=URL, json=json) for json in list_json]
            list_results = await asyncio.gather(*coros)

        results = {"inserted_row_count": 0}
        for r in list_results:
            results["inserted_row_count"] += r["inserted_row_count"]

        return results

    # Update Rows
    async def update_rows(self, table_name: str, updates: List[dict]):
        # updates = [{"row_id": xxx, "row": {"key": "value"}}, ...]
        METHOD = "PUT"
        URL = f"/dtable-server/api/v1/dtables/{self.base_token.dtable_uuid}/batch-update-rows/"

        # add select options if not exists
        _ = await self.add_select_options_if_not_exists(
            table_name=table_name, rows=[update["row"] for update in updates]
        )

        # divide chunk - [NOTE] 1000 rows까지만 됨
        UPDATE_LIMIT = 1000
        chunks = divide_chunks(updates, UPDATE_LIMIT)
        list_json = [{"table_name": table_name, "updates": chunk} for chunk in chunks]

        async with self.session_maker(token=self.base_token.access_token) as session:
            coros = [self.request(session=session, method=METHOD, url=URL, json=json) for json in list_json]
            list_results = await asyncio.gather(*coros)

        results = None
        for r in list_results:
            if isinstance(r, Exception):
                raise r
            if not r["success"]:
                results = r
                break
        else:
            results = r

        return results

    # (CUSTOM) Upsert Rows
    async def upsert_rows(self, table_name: str, rows: List[dict], key_column: str = None):
        if not key_column:
            table = await self.get_table(table_name=table_name)
            key_column = table.columns[0].name

        row_id_map = await self.generate_row_id_map(table_name=table_name, key_column=key_column)

        rows_to_update = list()
        rows_to_append = list()
        for row in rows:
            if row[key_column] in row_id_map:
                rows_to_update.append({"row_id": row_id_map[row[key_column]], "row": row})
            else:
                rows_to_append.append(row)

        results = dict()
        if rows_to_update:
            results_update = await self.update_rows(table_name=table_name, updates=rows_to_update)
            results.update({"update_rows": results_update})
        if rows_to_append:
            results_append = await self.append_rows(table_name=table_name, rows=rows_to_append)
            results.update({"append_rows": results_append})

        return results

    # Delete Rows
    async def delete_rows(self, table_name: str, row_ids: List[str]):
        METHOD = "DELETE"
        URL = f"/dtable-server/api/v1/dtables/{self.base_token.dtable_uuid}/batch-delete-rows/"

        json = {"table_name": table_name, "row_ids": row_ids}

        async with self.session_maker(token=self.base_token.access_token) as session:
            results = await self.request(session=session, method=METHOD, url=URL, json=json)

        return results

    # Rock Rows
    async def lock_rows(self, table_name: str, row_ids: List[str]):
        METHOD = "PUT"
        URL = f"/dtable-server/api/v1/dtables/{self.base_token.dtable_uuid}/lock-rows/"

        json = {"table_name": table_name, "row_ids": row_ids}

        async with self.session_maker(token=self.base_token.access_token) as session:
            results = await self.request(session=session, method=METHOD, url=URL, json=json)

        return results

    # Unrock Rows
    async def unlock_rows(self, table_name: str, row_ids: List[str]):
        METHOD = "PUT"
        URL = f"/dtable-server/api/v1/dtables/{self.base_token.dtable_uuid}/unlock-rows/"

        json = {"table_name": table_name, "row_ids": row_ids}

        async with self.session_maker(token=self.base_token.access_token) as session:
            results = await self.request(session=session, method=METHOD, url=URL, json=json)

        return results

    ################################################################
    # (CUSTOM) QUERY
    ################################################################
    # (CUSTOM) Query Key Map
    async def generate_row_id_map(self, table_name: str, key_column: str):
        results = await self.read_table(table_name=table_name, columns=["_id", key_column])
        return {r[key_column]: r["_id"] for r in results}

    # (CUSTOM) read_table
    async def read_table(
        self,
        table_name: str,
        columns: List[str] = None,
        modified_before: str = None,
        modified_after: str = None,
        offset: int = 0,
        limit: int = None,
        mtime: str = "_mtime",
        deserialize: bool = True,
    ) -> List[dict]:
        MAX_LIMIT = 10000
        OFFSET = 0

        # correct args
        table = PikaTable(table_name)
        if not columns:
            columns = ["*"]
        if not isinstance(columns, list):
            columns = [x.strip() for x in columns.split(",")]
        _limit = min(MAX_LIMIT, limit) if limit else limit
        _offset = offset if offset else OFFSET

        q = PikaQuery.from_(table).select(*columns)
        if modified_before:
            if isinstance(modified_before, datetime):
                modified_before = modified_before.isoformat(timespec="milliseconds")
            q = q.where(table[mtime] < modified_before)
        if modified_after:
            if isinstance(modified_after, datetime):
                modified_after = modified_after.isoformat(timespec="milliseconds")
            q = q.where(table[mtime] > modified_after)
        q = q.limit(_limit or MAX_LIMIT)

        # 1st hit
        rows = await self.list_rows_with_sql(sql=q.offset(_offset))

        # get all records
        if not limit or len(rows) < limit:
            while True:
                _offset += MAX_LIMIT
                _rows = await self.list_rows_with_sql(sql=q.offset(_offset))
                rows += _rows
                if len(_rows) < MAX_LIMIT:
                    break

        # to python data type
        if deserialize:
            deserializer = await self.generate_deserializer(table_name=table_name)
            try:
                rows = [deserializer(r) for r in rows]
            except Exception as ex:
                _msg = (
                    f"deserializer failed - group '{self.group_name}', base '{self.base_name}', table '{table_name}'"
                )
                logger.error(_msg)
                raise ex

        return rows

    # (CUSTOM) read table as DataFrame
    async def read_table_as_df(
        self,
        table_name: str,
        columns: List[str] = None,
        modified_before: str = None,
        modified_after: str = None,
        offset: int = 0,
        limit: int = None,
        mtime: str = "_mtime",
        deserialize: bool = True,
    ):
        rows = await self.read_table(
            table_name=table_name,
            columns=columns,
            modified_before=modified_before,
            modified_after=modified_after,
            offset=offset,
            limit=limit,
            mtime=mtime,
            deserialize=deserialize,
        )

        if not rows:
            return None
        tbl = pa.Table.from_pylist(rows).to_pandas()
        return tbl.set_index("_id", drop=True).rename_axis("row_id")

    # (CUSTOM) read view
    async def read_view(
        self,
        table_name: str,
        view_name: str,
        convert_link_id: bool = False,
        order_by: str = None,
        direction: str = "asc",
        start: int = 0,
        limit: int = None,
        deserialize: bool = True,
    ):
        rows = await self.list_rows(
            table_name=table_name,
            view_name=view_name,
            convert_link_id=convert_link_id,
            order_by=order_by,
            direction=direction,
            start=start,
            limit=limit,
        )

        # to python data type
        if deserialize:
            deserializer = await self.generate_deserializer(table_name=table_name)
            try:
                rows = [deserializer(r) for r in rows]
            except Exception as ex:
                _msg = f"deserializer failed - group '{self.group_name}', base '{self.base_name}', table '{table_name}', view '{view_name}'"
                logger.error(_msg)
                raise ex

        return rows

    # (CUSTOM) read view as DataFrame
    async def read_view_as_df(
        self,
        table_name: str,
        view_name: str,
        convert_link_id: bool = False,
        order_by: str = None,
        direction: str = "asc",
        start: int = 0,
        limit: int = None,
        deserialize: bool = True,
    ):
        rows = await self.read_view(
            table_name=table_name,
            view_name=view_name,
            convert_link_id=convert_link_id,
            order_by=order_by,
            direction=direction,
            start=start,
            limit=limit,
            deserialize=deserialize,
        )

        if not rows:
            return None
        tbl = pa.Table.from_pylist(rows).to_pandas()
        return tbl.set_index("_id", drop=True).rename_axis("row_id")

    # (CUSTOM) Generate Deserializer
    async def generate_deserializer(self, table_name):
        table = await self.get_table(table_name)
        users = await self.list_collaborators() if "collaborator" in [c.type for c in table.columns] else None
        return ToPythonDict(table=table, users=users)

    ################################################################
    # LINKS
    ################################################################

    ################################################################
    # FILES & IMAGES
    ################################################################

    ################################################################
    # TABLES
    ################################################################
    # Create New Table
    async def create_new_table(self, table_name: str, columns: List[Union[dict, SeaTableType]]):
        METHOD = "POST"
        URL = f"/dtable-server/api/v1/dtables/{self.base_token.dtable_uuid}/tables/"

        json = {
            "table_name": table_name,
            "columns": [c.seatable_schema() if isinstance(c, SeaTableType) else c for c in columns],
        }

        async with self.session_maker(token=self.base_token.access_token) as session:
            results = await self.request(session=session, method=METHOD, url=URL, json=json)

        return results

    # Rename Table
    async def rename_table(self, table_name: str, new_table_name: str):
        METHOD = "PUT"
        URL = f"/dtable-server/api/v1/dtables/{self.base_token.dtable_uuid}/tables/"

        json = {"table_name": table_name, "new_table_name": new_table_name}

        async with self.session_maker(token=self.base_token.access_token) as session:
            results = await self.request(session=session, method=METHOD, url=URL, json=json)

        return results

    # Delete Table
    async def delete_table(self, table_name: str):
        METHOD = "DELETE"
        URL = f"/dtable-server/api/v1/dtables/{self.base_token.dtable_uuid}/tables/"

        json = {"table_name": table_name}

        async with self.session_maker(token=self.base_token.access_token) as session:
            results = await self.request(session=session, method=METHOD, url=URL, json=json)

        return results

    # Duplicate Table
    async def duplicate_table(self, table_name: str, is_duplicate_records: bool = True):
        # rename table in a second step
        METHOD = "POST"
        URL = f"/dtable-server/api/v1/dtables/{self.base_token.dtable_uuid}/tables/duplicate-table/"

        json = {"table_name": table_name, "is_duplicate_records": is_duplicate_records}

        async with self.session_maker(token=self.base_token.access_token) as session:
            results = await self.request(session=session, method=METHOD, url=URL, json=json)

        return results

    ################################################################
    # IMPORT
    ################################################################
    # (DEPRECATED) Create Table from CSV

    # (DEPRECATED) Append Rows from CSV

    ################################################################
    # VIEWS
    ################################################################
    # List Views
    async def list_views(self, table_name: str, model: BaseModel = View):
        METHOD = "GET"
        URL = f"/dtable-server/api/v1/dtables/{self.base_token.dtable_uuid}/views/"
        ITEM = "views"

        async with self.session_maker(token=self.base_token.access_token) as session:
            response = await self.request(session=session, method=METHOD, url=URL, table_name=table_name)
            results = response[ITEM]

        if model:
            results = [model(**x) for x in results]

        return results

    # Create View
    async def create_view(
        self,
        table_name: str,
        name: str,
        type: str = "table",
        is_locked: bool = False,
        model: BaseModel = View,
    ):
        """
        type: "table" or "archive" (bigdata)
        """
        METHOD = "POST"
        URL = f"/dtable-server/api/v1/dtables/{self.base_token.dtable_uuid}/views/"

        json = {
            "name": name,
            "type": type,
            "is_locked": str(is_locked).lower(),
        }

        async with self.session_maker(token=self.base_token.access_token) as session:
            results = await self.request(
                session=session,
                method=METHOD,
                url=URL,
                json=json,
                table_name=table_name,
            )

        if model:
            results = model(**results)

        return results

    # Get View
    async def get_view(self, table_name: str, view_name: str, model: BaseModel = View):
        METHOD = "GET"
        URL = f"/dtable-server/api/v1/dtables/{self.base_token.dtable_uuid}/views/{view_name}/"

        async with self.session_maker(token=self.base_token.access_token) as session:
            results = await self.request(session=session, method=METHOD, url=URL, table_name=table_name)

        if model:
            results = model(**results)

        return results

    # (CUSTOM) Get View by ID
    async def get_view_by_id(self, table_id: str, view_id: str, model: BaseModel = View):
        table_name, view_name = await self.get_names_by_ids(table_id=table_id, view_id=view_id)

        return await self.get_view(table_name=table_name, view_name=view_name)

    # Update View
    # NOT TESTED!
    async def update_view(self, table_name: str, view_name: str, conf: Union[dict, BaseModel] = None):
        METHOD = "PUT"
        URL = f"/dtable-server/api/v1/dtables/{self.base_token.dtable_uuid}/views/{view_name}/"

        if isinstance(conf, BaseModel):
            conf = conf.dict()

        async with self.session_maker(token=self.base_token.access_token) as session:
            results = await self.request(
                session=session,
                method=METHOD,
                url=URL,
                json=conf,
                table_name=table_name,
            )

        return results

    # Delete View
    async def delete_view(self, table_name: str, view_name: str):
        METHOD = "DELETE"
        URL = f"/dtable-server/api/v1/dtables/{self.base_token.dtable_uuid}/views/{view_name}/"
        ITEM = "success"

        async with self.session_maker(token=self.base_token.access_token) as session:
            response = await self.request(session=session, method=METHOD, url=URL, table_name=table_name)
            results = response[ITEM]

        return results

    ################################################################
    # COLUMNS
    ################################################################
    async def add_select_options(
        self, table_name: str, column_name: str, options: List[SelectOption], model: BaseModel = None
    ):
        METHOD = "POST"
        URL = f"/dtable-server/api/v1/dtables/{self.base_token.dtable_uuid}/column-options/"

        json = {
            "table_name": table_name,
            "column": column_name,
            "options": [opt.dict(exclude_none=True) for opt in options],
        }

        async with self.session_maker(token=self.base_token.access_token) as session:
            results = await self.request(session=session, method=METHOD, url=URL, json=json)

        if model:
            results = model(**results)

        return results

    # (custom) add select options if not exists
    async def add_select_options_if_not_exists(self, table_name: str, rows: List[dict]):
        table = await self.get_table(table_name=table_name)
        columns_and_options = {
            c.name: [o["name"] for o in c.data["options"]]
            for c in table.columns
            if c.type in ["single-select", "multiple-select"]
        }

        if not columns_and_options:
            return

        options = {c: set([r.get(c) for r in rows]) for c in columns_and_options}
        options_to_add = dict()
        for column_name, column_options in options.items():
            for column_opt in column_options:
                if column_opt not in columns_and_options[column_name]:
                    if column_name not in options_to_add:
                        options_to_add[column_name] = list()
                    options_to_add[column_name].append(SelectOption(name=column_opt))

        coros = [
            self.add_select_options(table_name=table_name, column_name=column_name, options=options)
            for column_name, options in options_to_add.items()
        ]

        return await asyncio.gather(*coros)

    ################################################################
    # BIG DATA
    ################################################################

    ################################################################
    # ROW COMMENTS
    ################################################################
    async def list_row_comments(self, row_id: str):
        # NOT WORKING
        METHOD = "GET"
        URL = f"/dtable-server/api/v1/dtables/{self.base_token.dtable_uuid}/comments/"

        params = {"row_id": row_id}

        async with self.session_maker(token=self.base_token.access_token) as session:
            results = await self.request(session=session, method=METHOD, url=URL, **params)

        return results

    ################################################################
    # NOTIFICATION
    ################################################################

    ################################################################
    # ACTIVITIES & LOGS
    ################################################################
    # Get Base Activity Logs
    async def get_base_activity_log(self, page: int = 1, per_page: int = 25, model: BaseModel = BaseActivity):
        # rename table in a second step
        METHOD = "GET"
        URL = f"/dtable-server/api/v1/dtables/{self.base_token.dtable_uuid}/operations/"
        ITEM = "operations"

        params = {"page": page, "per_page": per_page}

        async with self.session_maker(token=self.base_token.access_token) as session:
            response = await self.request(session=session, method=METHOD, url=URL, **params)
            results = response[ITEM]

        if model:
            results = [model(**x) for x in results]

        return results

    # List Row Activities
    async def list_row_activities(self, row_id: str, page: int = 1, per_page: int = 25):
        # rename table in a second step
        METHOD = "GET"
        URL = f"/dtable-server/api/v1/dtables/{self.base_token.dtable_uuid}/activities/"

        params = {"row_id": row_id, "page": page, "per_page": per_page}

        async with self.session_maker(token=self.base_token.access_token) as session:
            results = await self.request(session=session, method=METHOD, url=URL, **params)

        return results

    # List Delete Operation Logs
    async def list_delete_operation_logs(self, op_type: str, page: int = 1, per_page: int = 25):
        """
        op_type
         delete_row
         delete_rows
         delete_table
         delete_column
        """
        # rename table in a second step
        METHOD = "GET"
        URL = f"/api/v2.1/dtables/{self.base_token.dtable_uuid}/delete-operation-logs/"

        params = {"op_type": op_type, "page": page, "per_page": per_page}

        async with self.session_maker(token=self.base_token.access_token) as session:
            results = await self.request(session=session, method=METHOD, url=URL, **params)

        return results

    # List Delete Rows
    async def list_delete_rows(self):
        # rename table in a second step
        METHOD = "GET"
        URL = f"/dtable-server/api/v1/dtables/{self.base_token.dtable_uuid}/deleted-rows/"

        async with self.session_maker(token=self.base_token.access_token) as session:
            results = await self.request(session=session, method=METHOD, url=URL)

        return results

    ################################################################
    # SNAPSHOTS
    ################################################################
    # TBD



import logging
from datetime import datetime
from typing import Any, List, Union

import pytz
from pydantic import BaseModel

from ..model import Table

logger = logging.getLogger(__name__)

SCHEMA_MAP = {
    "text": "text",
    "long_text": "long-text",
    "number": "number",
    "collaborator": "collaborator",
    "date": "date",
    "duration": "duration",
    "single_select": "single-select",
    "multiple_select": "multiple-select",
    "image": "image",
    "file": "file",
    "email": "email",
    "url": "url",
    "checkbox": "checkbox",
    "rating": "rating",
    "formula": "formula",
    "link": "link",
    "link_formula": "link-formula",
    "creator": "creator",
    "ctime": "ctime",
    "last_modifier": "last-modifier",
    "mtime": "mtime",
    "auto_number": "auto-number",
}

STR = {"column_type": "text"}
LONGTEXT = {"column_type": "long-text"}
INT = {
    "column_type": "number",
    "data": {"format": "number", "decimal": "dot", "thousands": "comma"},
}
FLOAT = {
    "column_type": "number",
    "data": {"format": "number", "decimal": "dot", "thousands": "comma"},
}
DATETIME = {"column_type": "date", "data": {"format": "YYYY-MM-DD HH:mm"}}
BOOL = {"column_type": "checkbox"}

BASE = {
    "table_name": "Base",
    "columns": [
        {"column_name": "base_uuid", **STR},  # "uuid"
        {"column_name": "group_name", **STR},  # "owner"
        {"column_name": "base_name", **STR},  # "name"
        {"column_name": "workspace_id", **INT},
        {"column_name": "created_at", **DATETIME},
        {"column_name": "updated_at", **DATETIME},
        {"column_name": "owner_deleted", **BOOL},
        {"column_name": "rows_count", **INT},
    ],
}

FASTO_USER = {
    "table_name": "FastoUser",
    "columns": {
        {"column_name": "ID", **STR},
        {"column_name": "EmailAddress", **STR},
        {"column_name": "Role", **STR},
        {"column_name": "UserCreated", **DATETIME},
        {"column_name": "LastLogin", **DATETIME},
        {"column_name": "IsActive", **BOOL},
        {"column_name": "Organization", **STR},
        {"column_name": "Division", **STR},
        {"column_name": "DeptCode", **STR},
        {"column_name": "DeptName", **STR},
    },
}



import asyncio
import logging
from typing import Callable

from plantable.client import AdminClient

from .conf import SEATABLE_PASSWORD, SEATABLE_URL, SEATABLE_USERNAME
from .ws_client import BaseWebsocketClient

logger = logging.getLogger(__file__)


# Seatable Producer
class Producer:
    def __init__(
        self,
        seatable_url: str = SEATABLE_URL,
        seatable_username: str = SEATABLE_USERNAME,
        seatable_password: str = SEATABLE_PASSWORD,
        handler: Callable = None,
        wait_for: float = 2.0,
        ws_reconnection_delay=3,
        ws_reconnection_attempts=40,
    ):
        self.seatable_url = seatable_url
        self.seatable_username = seatable_username
        self.seatable_password = seatable_password
        self.handler = handler
        self.wait_for = wait_for

        self.ws_reconnection_delay = ws_reconnection_delay
        self.ws_reconnection_attempts = ws_reconnection_attempts

        self.client = AdminClient(
            seatable_url=self.seatable_url,
            seatable_username=self.seatable_username,
            seatable_password=self.seatable_password,
        )

        self.tasks = dict()
        self.clients = dict()

    # run
    async def run(self, debug: bool = False):
        try:
            await self.watch(debug=debug)
        except KeyboardInterrupt:
            logger.error("Canelled!")
            for group_id, bases in self.tasks.items():
                for base_id, task in bases.items():
                    await self.cancel_task(task)
                    _msg = f"Removed: Group {group_id}, Base {base_id}"
                    print(_msg)
            return

    # list views with name
    async def watch(self, debug: bool = False):
        try:
            while True:
                tasks = dict()
                groups = await self.client.list_groups()
                for group in groups:
                    if group.id not in tasks:
                        tasks[group.id] = dict()
                    bases = await self.client.list_group_bases(group.name)
                    for base in bases:
                        if base.id not in tasks[group.id]:
                            tasks[group.id].update({base.id: base.name})

                # remove deleted bases
                empty_groups = list()
                for group_id, bases in self.tasks.items():
                    empty_bases = list()
                    if not bases:
                        empty_groups.append(group_id)
                    if group_id not in tasks:
                        for base_id, task in bases.items():
                            await self.cancel_task(task)
                            _msg = f"Removed: Group {group_id}, Base {base_id}"
                            print(_msg)
                            empty_bases.append(base_id)
                    for base_id, task in bases.items():
                        if base_id not in tasks[group_id]:
                            await self.cancel_task(task)
                            _msg = f"Removed: {group_id}, Base {base_id}"
                            print(_msg)
                            empty_bases.append(base_id)
                    for empty_base in empty_bases:
                        self.tasks[group_id].pop(empty_base)
                for empty_group in empty_groups:
                    self.tasks.pop(empty_group)

                # update tasks
                for group_id, bases in tasks.items():
                    if group_id not in self.tasks:
                        self.tasks[group_id] = dict()
                    for base_id, base_name in bases.items():
                        if base_id not in self.tasks[group_id] or self.tasks[group_id][base_id].done():
                            try:
                                self.tasks[group_id][base_id] = asyncio.create_task(
                                    self.run_websocket(group_id=group_id, base_name=base_name)
                                )
                                _msg = f"Registered: Group {group_id}, Base {base_id}"
                                print(_msg)
                            except Exception as ex:
                                _msg = f"FAILED: create websocket and run - Group {group_id}, Base {base_id} (exceoption: {ex})"
                                logger.warning(_msg)
                if debug:
                    break
                await asyncio.sleep(self.wait_for)
        except asyncio.CancelledError:
            return

    # Create Websocket
    async def run_websocket(self, group_id: int, base_name: str):
        client = BaseWebsocketClient(
            group_name_or_id=group_id,
            base_name=base_name,
            seatable_url=self.seatable_url,
            seatable_username=self.seatable_username,
            seatable_password=self.seatable_password,
            handler=self.handler,
            reconnection_delay=self.ws_reconnection_attempts,
            reconnection_attempts=self.ws_reconnection_attempts,
        )
        try:
            await client.run()
        except KeyboardInterrupt:
            await client.disconnect()

    @staticmethod
    async def cancel_task(task):
        task.cancel()
        while not task.done():
            await asyncio.sleep(0.1)
        return



################################################################
# Row Events
################################################################
# Insert Row
sample_insert_row = {
    "op_type": "insert_row",
    "table_id": "wMtQ",
    "row_id": "a4MUb9F2Sb6erVSK-WWCkA",
    "row_insert_position": "insert_below",
    "row_data": {
        "_id": "FPFMs5blRR-SpLV1sk676g",
        "_participants": [],
        "_creator": "2926d3fa3a364558bac8a550811dbe0e@auth.local",
        "_ctime": "2023-08-20T04:59:21.673+00:00",
        "_last_modifier": "2926d3fa3a364558bac8a550811dbe0e@auth.local",
        "_mtime": "2023-08-20T04:59:21.673+00:00",
    },
    "links_data": {},
    "key_auto_number_config": {},
}

# Insert Rows
sample_insert_rows = {
    "op_type": "insert_rows",
    "table_id": "wMtQ",
    "row_ids": ["FPFMs5blRR-SpLV1sk676g", "NLEm269CSaO3gy7Cm2ZrvQ"],
    "row_insert_position": "insert_below",
    "row_datas": [
        {
            "_id": "NLEm269CSaO3gy7Cm2ZrvQ",
            "_participants": [],
            "_creator": "2926d3fa3a364558bac8a550811dbe0e@auth.local",
            "_ctime": "2023-08-20T04:59:35.211+00:00",
            "_last_modifier": "2926d3fa3a364558bac8a550811dbe0e@auth.local",
            "_mtime": "2023-08-20T04:59:35.211+00:00",
            "0000": "",
        },
        {
            "_id": "I0n3PtjORke3ubWq5B54GQ",
            "_participants": [],
            "_creator": "2926d3fa3a364558bac8a550811dbe0e@auth.local",
            "_ctime": "2023-08-20T04:59:35.211+00:00",
            "_last_modifier": "2926d3fa3a364558bac8a550811dbe0e@auth.local",
            "_mtime": "2023-08-20T04:59:35.211+00:00",
            "0000": "",
        },
    ],
    "links_data": {},
    "key_auto_number_config": {},
}


# Append Rows (Append Row는 없음)
sample_append_rows = {
    "op_type": "append_rows",
    "table_id": "wMtQ",
    "row_datas": [
        {
            "0000": "w.cho@cj.net",
            "HkRD": "조우진",
            "sAd6": "DT플랫폼팀",
            "WD7B": "",
            "_last_modifier": "2926d3fa3a364558bac8a550811dbe0e@auth.local",
            "_creator": "2926d3fa3a364558bac8a550811dbe0e@auth.local",
            "_id": "F7ze485HTAKfrZmCZYqb1g",
            "_ctime": "2023-08-20T05:15:42.105+00:00",
            "_mtime": "2023-08-20T05:15:42.105+00:00",
        },
        {
            "0000": "w.cho@cj.net",
            "HkRD": "a",
            "sAd6": "",
            "WD7B": "",
            "_last_modifier": "2926d3fa3a364558bac8a550811dbe0e@auth.local",
            "_creator": "2926d3fa3a364558bac8a550811dbe0e@auth.local",
            "_id": "ONBSOt73RieaD7O0m4LiaA",
            "_ctime": "2023-08-20T05:15:42.106+00:00",
            "_mtime": "2023-08-20T05:15:42.106+00:00",
        },
    ],
    "key_auto_number_config": {},
}

# Modify Row
sample_modify_row = {
    "op_type": "modify_row",
    "table_id": "wMtQ",
    "row_id": "FPFMs5blRR-SpLV1sk676g",
    "updated": {
        "HkRD": "a",
        "_last_modifier": "2926d3fa3a364558bac8a550811dbe0e@auth.local",
    },
    "old_row": {},
}

# Modify Rows
sample_modify_rows = {
    "op_type": "modify_rows",
    "table_id": "wMtQ",
    "row_ids": ["X-jLvmZQRq-lMgp6X1zPgw", "a4MUb9F2Sb6erVSK-WWCkA"],
    "updated": {
        "X-jLvmZQRq-lMgp6X1zPgw": {
            "HkRD": "a",
            "_last_modifier": "2926d3fa3a364558bac8a550811dbe0e@auth.local",
        },
        "a4MUb9F2Sb6erVSK-WWCkA": {
            "HkRD": "a",
            "_last_modifier": "2926d3fa3a364558bac8a550811dbe0e@auth.local",
        },
    },
    "old_rows": {
        "X-jLvmZQRq-lMgp6X1zPgw": {"HkRD": None},
        "a4MUb9F2Sb6erVSK-WWCkA": {"HkRD": None},
    },
}

# Delete Row
sample_delete_row = {
    "op_type": "delete_row",
    "table_id": "wMtQ",
    "row_id": "bH1cdZboS7WOg_qpjUkOXw",
    "deleted_row": {
        "_id": "bH1cdZboS7WOg_qpjUkOXw",
        "_participants": [],
        "_creator": "2926d3fa3a364558bac8a550811dbe0e@auth.local",
        "_ctime": "2023-08-20T04:37:21.579+00:00",
        "_last_modifier": "2926d3fa3a364558bac8a550811dbe0e@auth.local",
        "_mtime": "2023-08-20T04:37:21.587+00:00",
        "0000": "",
    },
    "upper_row_id": "JmsWPdqgT4irjvxAyKBxlA",
    "deleted_links_data": {},
}

# Delete Rows
sample_delete_rows = {
    "op_type": "delete_rows",
    "table_id": "wMtQ",
    "row_ids": ["NwkAMFWgTzO6A8ugLyMzMQ", "JmsWPdqgT4irjvxAyKBxlA"],
    "deleted_rows": [
        {
            "_id": "NwkAMFWgTzO6A8ugLyMzMQ",
            "_participants": [],
            "_creator": "2926d3fa3a364558bac8a550811dbe0e@auth.local",
            "_ctime": "2023-08-20T04:33:50.072+00:00",
            "_last_modifier": "2926d3fa3a364558bac8a550811dbe0e@auth.local",
            "_mtime": "2023-08-20T04:33:50.072+00:00",
        },
        {
            "_id": "JmsWPdqgT4irjvxAyKBxlA",
            "_participants": [],
            "_creator": "2926d3fa3a364558bac8a550811dbe0e@auth.local",
            "_ctime": "2023-08-20T04:37:21.579+00:00",
            "_last_modifier": "2926d3fa3a364558bac8a550811dbe0e@auth.local",
            "_mtime": "2023-08-20T04:37:21.587+00:00",
            "0000": "",
        },
    ],
    "upper_row_ids": ["H2Ib-4kxRWmlxIo2ft0IGQ", "X-jLvmZQRq-lMgp6X1zPgw"],
    "deleted_links_data": {},
}


# Update Rows Links
sample_update_rows_links = {
    "op_type": "update_rows_links",
    "link_id": "t4hs",
    "table_id": "hNbE",
    "other_table_id": "ric6",
    "row_id_list": ["VHUU0Fx3T9qagNXpDXiQnw", "QFwE8A2fT92ZGUt_tK3s1Q"],
    "other_rows_ids_map": {
        "VHUU0Fx3T9qagNXpDXiQnw": ["Z6sr4Z5wQ_6b1AMRxQzZgg"],
        "QFwE8A2fT92ZGUt_tK3s1Q": ["Z6sr4Z5wQ_6b1AMRxQzZgg"],
    },
    "old_other_rows_ids_map": {"VHUU0Fx3T9qagNXpDXiQnw": [], "QFwE8A2fT92ZGUt_tK3s1Q": []},
}

# Archive Rows
sample_archive_rows = {"op_type": "archive_rows", "table_id": "04Af", "row_ids": ["NGsQGjTlSN6_qbAiAi1Skw"]}


################################################################
# Table Events
################################################################
# Insert Table
sample_insert_table = {
    "op_type": "insert_table",
    "table_data": {
        "_id": "D4S8",
        "name": "Table3",
        "is_header_locked": False,
        "header_settings": {},
        "summary_configs": {},
        "columns": [
            {
                "key": "0000",
                "type": "text",
                "name": "Name",
                "editable": True,
                "width": 200,
                "resizable": True,
                "draggable": True,
                "data": None,
                "permission_type": "",
                "permitted_users": [],
                "edit_metadata_permission_type": "",
                "edit_metadata_permitted_users": [],
                "description": None,
            }
        ],
        "rows": [],
        "view_structure": {"folders": [], "view_ids": ["0000"]},
        "views": [
            {
                "_id": "0000",
                "name": "Default View",
                "type": "table",
                "private_for": None,
                "is_locked": False,
                "row_height": "default",
                "filter_conjunction": "And",
                "filters": [],
                "sorts": [],
                "groupbys": [],
                "colorbys": {},
                "hidden_columns": [],
                "rows": [],
                "formula_rows": {},
                "link_rows": {},
                "summaries": {},
                "colors": {},
                "column_colors": {},
                "groups": [],
            }
        ],
        "id_row_map": {},
    },
}

# Rename Table
sample_rename_table = {
    "op_type": "rename_table",
    "table_id": "0000",
    "table_name": "Table_Renamed",
}

# Delete Table
sample_delete_table = {
    "op_type": "delete_table",
    "table_id": "CzsA",
    "table_name": "Table3(copy)",
    "deleted_table": {
        "_id": "CzsA",
        "name": "Table3(copy)",
        "is_header_locked": False,
        "columns": [
            {
                "key": "0000",
                "type": "text",
                "name": "Name",
                "editable": True,
                "width": 200,
                "resizable": True,
                "draggable": True,
                "data": None,
                "permission_type": "",
                "permitted_users": [],
                "edit_metadata_permission_type": "",
                "edit_metadata_permitted_users": [],
                "description": None,
                "editor": {"key": None, "ref": None, "props": {}, "_owner": None},
                "formatter": {"key": None, "ref": None, "props": {}, "_owner": None},
            }
        ],
        "rows": [],
        "id_row_map": {},
        "view_structure": {"folders": [], "view_ids": ["0000"]},
        "views": [
            {
                "_id": "0000",
                "name": "Default View",
                "type": "table",
                "private_for": None,
                "is_locked": False,
                "row_height": "default",
                "filter_conjunction": "And",
                "filters": [],
                "sorts": [],
                "groupbys": [],
                "colorbys": {},
                "hidden_columns": [],
                "rows": [],
                "formula_rows": {},
                "link_rows": {},
                "summaries": {},
                "colors": {},
                "column_colors": {},
                "groups": [],
            }
        ],
        "header_settings": {},
    },
}

sample_header_lock = {
    "op_type": "modify_header_lock",
    "table_id": "hNbE",
    "is_header_locked": False,
}


################################################################
# Column Events
################################################################
# Insert Column [DONE]
sample_insert_column = {
    "op_type": "insert_column",
    "table_id": "wMtQ",
    "column_key": "_last_modifier",
    "column_data": {
        "key": "4xVF",
        "type": "text",
        "name": "hello",
        "editable": True,
        "width": 200,
        "resizable": True,
        "draggable": True,
        "data": {
            "enable_fill_default_value": False,
            "enable_check_format": False,
            "format_specification_value": None,
            "default_value": "",
            "format_check_type": "custom_format",
        },
        "permission_type": "",
        "permitted_users": [],
        "edit_metadata_permission_type": "",
        "edit_metadata_permitted_users": [],
        "description": None,
    },
    "view_id": "0000",
    "rows_datas": [],
}

# Delete Column [DONE]
sample_delete_column = {
    "op_type": "delete_column",
    "table_id": "wMtQ",
    "column_key": "4xVF",
    "old_column": {
        "rowType": "header",
        "key": "4xVF",
        "type": "text",
        "name": "hello",
        "editable": True,
        "width": 200,
        "resizable": True,
        "draggable": True,
        "data": {
            "enable_fill_default_value": False,
            "enable_check_format": False,
            "format_specification_value": None,
            "default_value": "",
            "format_check_type": "custom_format",
        },
        "permission_type": "",
        "permitted_users": [],
        "edit_metadata_permission_type": "",
        "edit_metadata_permitted_users": [],
        "description": None,
        "editor": {"key": None, "ref": None, "props": {}, "_owner": None},
        "formatter": {"key": None, "ref": None, "props": {}, "_owner": None},
        "idx": 6,
        "left": 1080,
        "last_frozen": False,
    },
    "upper_column_key": "_last_modifier",
}

# Rename Column [Done]
sample_rename_column = {
    "op_type": "rename_column",
    "table_id": "wMtQ",
    "column_key": "WD7B",
    "new_column_name": "what",
    "old_column_name": "guest_added",
}

# Update Column Description [Done]
sample_update_column_description = {
    "op_type": "update_column_description",
    "table_id": "wMtQ",
    "column_key": "WD7B",
    "column_description": "description here",
}


# Modify Column Type [Done]
sample_modify_column_type = {
    "op_type": "modify_column_type",
    "table_id": "wMtQ",
    "column_key": "WD7B",
    "new_column": {
        "key": "WD7B",
        "type": "email",
        "name": "guest_added",
        "editable": True,
        "width": 200,
        "resizable": True,
        "draggable": True,
        "data": None,
        "permission_type": "",
        "permitted_users": [],
        "edit_metadata_permission_type": "",
        "edit_metadata_permitted_users": [],
        "description": None,
        "editor": {"key": None, "ref": None, "props": {}, "_owner": None},
        "formatter": {"key": None, "ref": None, "props": {}, "_owner": None},
    },
    "old_column": {
        "key": "WD7B",
        "type": "text",
        "name": "guest_added",
        "editable": True,
        "width": 200,
        "resizable": True,
        "draggable": True,
        "data": {
            "enable_fill_default_value": False,
            "enable_check_format": False,
            "format_specification_value": None,
            "default_value": "",
            "format_check_type": "custom_format",
        },
        "permission_type": "",
        "permitted_users": [],
        "edit_metadata_permission_type": "",
        "edit_metadata_permitted_users": [],
        "description": None,
        "editor": {"key": None, "ref": None, "props": {}, "_owner": None},
        "formatter": {"key": None, "ref": None, "props": {}, "_owner": None},
    },
    "new_rows_data": [
        {"_id": "H2Ib-4kxRWmlxIo2ft0IGQ", "WD7B": None},
        {"_id": "X-jLvmZQRq-lMgp6X1zPgw", "WD7B": None},
        {"_id": "a4MUb9F2Sb6erVSK-WWCkA", "WD7B": None},
    ],
    "old_rows_data": [
        {"_id": "H2Ib-4kxRWmlxIo2ft0IGQ", "WD7B": None},
        {"_id": "X-jLvmZQRq-lMgp6X1zPgw"},
        {"_id": "a4MUb9F2Sb6erVSK-WWCkA"},
    ],
}


# Modify Column Permission
sample_modify_column_permission = {
    "op_type": "modify_column_permission",
    "table_id": "wMtQ",
    "column_key": "WD7B",
    "new_column_permission": {"permission_type": "admins", "permitted_users": []},
    "old_column_permission": {"permission_type": "", "permitted_users": []},
}


# Modify Column Metadata Permission
sample_modify_column_metadata_permission = {
    "op_type": "modify_column_metadata_permission",
    "table_id": "wMtQ",
    "column_key": "WD7B",
    "new_column_permission": {
        "edit_metadata_permission_type": "admins",
        "edit_metadata_permitted_users": [],
    },
    "old_column_permission": {
        "edit_metadata_permission_type": "",
        "edit_metadata_permitted_users": [],
    },
}

# Update Column Colorbys
sample_update_column_colorybys = {
    "op_type": "update_column_colorbys",
    "table_id": "wMtQ",
    "column_key": "WD7B",
    "colorbys": {
        "type": "by_numeric_range",
        "range_settings": {
            "color_type": "color_gradation_1",
            "is_custom_start_value": True,
            "is_custom_end_value": False,
            "start_value": "",
            "end_value": "",
        },
    },
}


################################################################
# View Events
################################################################
# Insert View
sample_insert_view = {
    "op_type": "insert_view",
    "table_id": "wMtQ",
    "view_data": {
        "_id": "ht2Q",
        "name": "test",
        "type": "table",
        "private_for": None,
        "is_locked": False,
        "row_height": "default",
        "filter_conjunction": "And",
        "filters": [],
        "sorts": [],
        "groupbys": [],
        "colorbys": {},
        "hidden_columns": [],
        "rows": [],
        "formula_rows": {},
        "link_rows": {},
        "summaries": {},
        "colors": {},
        "column_colors": {},
        "groups": [],
    },
    "view_folder_id": None,
}

# Delete View
sample_delete_view = {
    "op_type": "delete_view",
    "table_id": "0000",
    "view_id": "761d",
    "view_folder_id": None,
    "view_name": "abc",
}

# Rename View
sample_rename_view = {
    "op_type": "rename_view",
    "table_id": "wMtQ",
    "view_id": "ht2Q",
    "view_name": "test_again",
}

# Move View
sample_move_view = {
    "op_type": "move_view",
    "table_id": "04Af",
    "moved_view_id": "9kc1",
    "target_view_id": "0000",
    "source_view_folder_id": None,
    "target_view_folder_id": None,
    "move_position": "move_above",
    "moved_view_name": "Raw",
}

# Modify View Type (Big Data)
sample_modify_view_type = {
    "op_type": "modify_view_type",
    "table_id": "hNbE",
    "view_id": "0oPc",
    "view_type": "archive",
}

# Modify View Lock
sample_modify_view_lock = {
    "op_type": "modify_view_lock",
    "table_id": "wMtQ",
    "view_id": "ht2Q",
    "is_locked": True,
}

# Modify Filter
sample_modify_filters = {
    "op_type": "modify_filters",
    "table_id": "wMtQ",
    "view_id": "ht2Q",
    "filters": [{"column_key": "0000", "filter_predicate": "contains", "filter_term": "woojin"}],
    "filter_conjunction": "And",
}

# Modify Sorts
sample_modify_sorts = {
    "op_type": "modify_sorts",
    "table_id": "wMtQ",
    "view_id": "ht2Q",
    "sorts": [{"column_key": "HkRD", "sort_type": "up"}],
}

# Modify Groupbys
sample_modify_groupbys = {
    "op_type": "modify_groupbys",
    "table_id": "wMtQ",
    "view_id": "ht2Q",
    "groupbys": [{"column_key": "HkRD", "sort_type": "up", "count_type": ""}],
}

# Modify Hidden Columns
sample_modify_hidden_columns = {
    "op_type": "modify_hidden_columns",
    "table_id": "wMtQ",
    "view_id": "ht2Q",
    "hidden_columns": ["WD7B"],
}

# Modify Row Color (Rows in Column)
sample_modify_row_color = {
    "op_type": "modify_row_color",
    "table_id": "wMtQ",
    "view_id": "ht2Q",
    "colorbys": {"type": "by_duplicate_values", "color_by_duplicate_column_keys": ["HkRD", "sAd6"]},
}

# Modify Row Height
sample_modify_row_height = {
    "op_type": "modify_row_height",
    "table_id": "wMtQ",
    "view_id": "ht2Q",
    "row_height": "quadruple",
}



from pydantic import BaseModel, root_validator

from .model import Event
from .const import (
    OP_INSERT_ROW,
    OP_INSERT_ROWS,
    OP_DELETE_ROW,
    OP_APPEND_ROW,
    OP_APPEND_ROWS,
    OP_DELETE_ROWS,
    OP_MODIFY_ROW,
    OP_MODIFY_ROWS,
    OP_UPDATE_ROW_LINKS,
    OP_UPDATE_ROWS_LINKS,
    OP_ARCHIVE_ROW,
    OP_ARCHIVE_ROWS,
)


################################################################
# Models
################################################################
# Row Event
class RowEvent(Event):
    op_type: str
    table_id: str
    row_id: str = None  # insert 기준 row의 id, append_rows에만 없음


# Insert Row
class InsertRow(RowEvent):
    anchor_row_id: str  # Insert의 기준 row
    row_insert_position: str  # 기준 row에 대한 insert position
    row_data: dict
    links_data: dict = None
    key_auto_number_config: dict

    @root_validator(pre=True)
    def adjust_row_id(cls, values):
        values["anchor_row_id"] = values["row_id"]
        values.update({"row_id": values["row_data"]["_id"]})
        return values


# Append Row
class AppendRow(RowEvent):
    row_data: dict
    key_auto_number_config: dict


# Modify Row
class ModifyRow(RowEvent):
    updated: dict
    old_row: dict


# Delete Row
class DeleteRow(RowEvent):
    deleted_row: dict
    upper_row_id: str = None
    deleted_links_data: dict


# Update Row Link
class UpdateRowLinks(RowEvent):
    other_table_id: str
    other_row_ids_map: dict
    old_other_row_ids_map: dict


# ArchiveRows
class ArchiveRow(RowEvent):
    pass


################################################################
# Parser
################################################################
def row_event_parser(data):
    op_type = data["op_type"]

    # INSERT
    if op_type == OP_INSERT_ROW:
        return [InsertRow(**data)]

    if op_type == OP_INSERT_ROWS:
        return [
            InsertRow(
                op_type=OP_INSERT_ROW,
                table_id=data["table_id"],
                row_id=row_id,
                row_insert_position=data["row_insert_position"],
                row_data=row_data,
                links_data=data["links_data"],
                key_auto_number_config=data["key_auto_number_config"],
            )
            for row_id, row_data in zip(data["row_ids"], data["row_datas"])
        ]

    # APPEND
    if op_type == OP_APPEND_ROWS:
        return [
            AppendRow(
                op_type=OP_APPEND_ROW,  # [NOTE] OP_APPEND_ROW! not OP_APPEND_ROWS
                table_id=data["table_id"],
                row_data=row_data,
                key_auto_number_config=data["key_auto_number_config"],
            )
            for row_data in data["row_datas"]
        ]

    # MODIFY
    if op_type == OP_MODIFY_ROW:
        return [ModifyRow(**data)]

    if op_type == OP_MODIFY_ROWS:
        return [
            ModifyRow(
                op_type=OP_MODIFY_ROW,
                table_id=data["table_id"],
                row_id=row_id,
                updated=updated,
                old_row=old_row,
            )
            for row_id, updated, old_row in zip(
                data["row_ids"],
                [v for _, v in data["updated"].items()],
                [v for _, v in data["old_rows"].items()],
            )
        ]

    # DELETE
    if op_type == OP_DELETE_ROW:
        return [DeleteRow(**data)]

    if op_type == OP_DELETE_ROWS:
        return [
            DeleteRow(
                op_type=OP_DELETE_ROW,
                table_id=data["table_id"],
                row_id=row_id,
                deleted_row=deleted_row,
                upper_row_id=upper_row_id,
                deleted_links_data=data["deleted_links_data"],
            )
            for row_id, deleted_row, upper_row_id in zip(data["row_ids"], data["deleted_rows"], data["upper_row_ids"])
        ]

    # UPDATE ROW LINK
    if op_type == OP_UPDATE_ROWS_LINKS:
        return [
            UpdateRowLinks(
                op_type=OP_UPDATE_ROW_LINKS,
                table_id=data["table_id"],
                other_table_id=data["other_table_id"],
                row_id=row_id,
                other_row_ids_map={new_key: new_value},
                old_other_row_ids_map={old_key: old_value},
            )
            for row_id, (new_key, new_value), (old_key, old_value) in zip(
                data["row_id_list"],
                data["other_rows_ids_map"].items(),
                data["old_other_rows_ids_map"].items(),
            )
        ]

    # ARCHIVE ROWS
    if op_type == OP_ARCHIVE_ROWS:
        return [
            ArchiveRow(op_type=OP_ARCHIVE_ROW, table_id=data["table_id"], row_id=row_id) for row_id in data["row_ids"]
        ]

    _msg = f"Row Parser - Unknown op_type '{op_type}'!"
    raise KeyError(_msg)



# ROWS
OP_INSERT_ROW = "insert_row"
OP_INSERT_ROWS = "insert_rows"
OP_APPEND_ROW = "append_row"
OP_APPEND_ROWS = "append_rows"
OP_MODIFY_ROW = "modify_row"
OP_MODIFY_ROWS = "modify_rows"
OP_DELETE_ROW = "delete_row"
OP_DELETE_ROWS = "delete_rows"
OP_UPDATE_ROW_LINKS = "update_row_links"
OP_UPDATE_ROWS_LINKS = "update_rows_links"
OP_ARCHIVE_ROW = "archive_row"
OP_ARCHIVE_ROWS = "archive_rows"

ROW_EVENT_CAT_NAME = "row_event"
ROW_EVENTS = [
    OP_INSERT_ROW,
    OP_INSERT_ROWS,
    OP_APPEND_ROW,
    OP_APPEND_ROWS,
    OP_MODIFY_ROW,
    OP_MODIFY_ROWS,
    OP_DELETE_ROW,
    OP_DELETE_ROWS,
    OP_UPDATE_ROW_LINKS,
    OP_UPDATE_ROWS_LINKS,
    OP_ARCHIVE_ROW,
    OP_ARCHIVE_ROWS,
]


# COLUMNS
OP_INSERT_COLUMN = "insert_column"
OP_DELETE_COLUMN = "delete_column"
OP_RENAME_COLUMN = "rename_column"
OP_UPDATE_COLUMN_DESCRIPTION = "update_column_description"
OP_UPDATE_COLUMN_COLORBYS = "update_column_colorbys"
OP_MODIFY_COLUMN_TYPE = "modify_column_type"
OP_MODIFY_COLUMN_PERMISSION = "modify_column_permission"
OP_MODIFY_COLUMN_METADATA_PERMISSION = "modify_column_metadata_permission"

COLUMN_EVENT_CAT_NAME = "column_event"
COLUMN_EVENTS = [
    OP_INSERT_COLUMN,
    OP_DELETE_COLUMN,
    OP_RENAME_COLUMN,
    OP_UPDATE_COLUMN_DESCRIPTION,
    OP_UPDATE_COLUMN_COLORBYS,
    OP_MODIFY_COLUMN_TYPE,
    OP_MODIFY_COLUMN_PERMISSION,
    OP_MODIFY_COLUMN_METADATA_PERMISSION,
]


# VIEW
OP_INSERT_VIEW = "insert_view"
OP_DELETE_VIEW = "delete_view"
OP_RENAME_VIEW = "rename_view"
OP_MODIFY_VIEW_TYPE = "modify_view_type"
OP_MODIFY_VIEW_LOCK = "modify_view_lock"
OP_MODIFY_FILTERS = "modify_filters"
OP_MODIFY_SORTS = "modify_sorts"
OP_MODIFY_GROUPBYS = "modify_groupbys"
OP_MODIFY_HIDDEN_COLUMNS = "modify_hidden_columns"
OP_MODIFY_ROW_COLOR = "modify_row_color"
OP_MODIFY_ROW_HEIGHT = "modify_row_height"
OP_MOVE_VIEW = "move_view"

VIEW_EVENT_CAT_NAME = "view_event"
VIEW_EVENTS = [
    OP_INSERT_VIEW,
    OP_DELETE_VIEW,
    OP_RENAME_VIEW,
    OP_MODIFY_VIEW_TYPE,
    OP_MODIFY_VIEW_LOCK,
    OP_MODIFY_FILTERS,
    OP_MODIFY_SORTS,
    OP_MODIFY_GROUPBYS,
    OP_MODIFY_HIDDEN_COLUMNS,
    OP_MODIFY_ROW_COLOR,
    OP_MODIFY_ROW_HEIGHT,
    OP_MOVE_VIEW,
]

# TABLE
OP_INSERT_TABLE = "insert_table"
OP_RENAME_TABLE = "rename_table"
OP_DELETE_TABLE = "delete_table"
OP_MODIFY_HEADER_LOCK = "modify_header_lock"

TABLE_EVENT_CAT_NAME = "table_event"
TABLE_EVENTS = [
    OP_INSERT_TABLE,
    OP_RENAME_TABLE,
    OP_DELETE_TABLE,
    OP_MODIFY_HEADER_LOCK,
]

# PARQUET OVERWRITE CASE
PARQUET_OVERWRITE_EVENTS = [
    # COLUMNS
    OP_INSERT_COLUMN,
    OP_DELETE_COLUMN,
    OP_RENAME_COLUMN,
    # ROWS
    OP_INSERT_ROW,
    OP_INSERT_ROWS,
    OP_APPEND_ROW,
    OP_APPEND_ROWS,
    OP_MODIFY_ROW,
    OP_MODIFY_ROWS,
    OP_DELETE_ROW,
    OP_DELETE_ROWS,
    # TABLE
    OP_RENAME_TABLE,
    # VIEW
    OP_RENAME_VIEW,
    OP_MODIFY_FILTERS,
    OP_MODIFY_SORTS,
    OP_MODIFY_HIDDEN_COLUMNS,
]


# Helper
def get_event_cat_name(op_type: str):
    if op_type in ROW_EVENTS:
        return ROW_EVENT_CAT_NAME
    if op_type in COLUMN_EVENTS:
        return COLUMN_EVENT_CAT_NAME
    if op_type in VIEW_EVENTS:
        return VIEW_EVENT_CAT_NAME
    if op_type in TABLE_EVENTS:
        return TABLE_EVENT_CAT_NAME
    _msg = f"route_event_type - unknown op_type '{op_type}'"
    raise KeyError(_msg)



from typing import Any, List

from pydantic import BaseModel, Field, root_validator

from .const import OP_DELETE_TABLE, OP_INSERT_TABLE, OP_RENAME_TABLE, OP_MODIFY_HEADER_LOCK
from .model import Column, Event, Table, View


################################################################
# Model
################################################################
# Table Event
class TableEvent(Event):
    op_type: str
    table_id: str


# Insert Table
class InsertTable(TableEvent):
    table_data: Table

    @root_validator(pre=True)
    def update_table_id_from_table_data(cls, values):
        values.update({"table_id": values["table_data"]["_id"]})
        return values


# Rename Table
class RenameTable(TableEvent):
    table_name: str


# Delete Table
class DeleteTable(TableEvent):
    table_name: str = None
    deleted_table: Table


# Modify Header Lock
class ModifyHeaderLock(TableEvent):
    is_header_locked: bool


################################################################
# Table Parser
################################################################
def table_event_parser(data):
    op_type = data["op_type"]

    if op_type == OP_INSERT_TABLE:
        return [InsertTable(**data)]

    if op_type == OP_RENAME_TABLE:
        return [RenameTable(**data)]

    if op_type == OP_DELETE_TABLE:
        return [DeleteTable(**data)]

    if op_type == OP_MODIFY_HEADER_LOCK:
        return [ModifyHeaderLock(**data)]

    _msg = f"table_event_parser - unknown op_type '{op_type}'!"
    raise KeyError(_msg)



import os

import pytz

TZ_INFO = os.getenv("TZ", "Asia/Seoul")
TZ = pytz.timezone(TZ_INFO) if TZ_INFO else pytz.UTC
DT_FMT = "%Y-%m-%dT%H:%M:%S.%f%z"

################################################################
# SeaTable
################################################################
# System Fields
SYSTEM_FIELDS = {
    "_id": {"column_type": "text"},
    "_locked": {"column_type": "checkbox"},
    "_locked_by": {"column_type": "text"},
    "_archived": {"column_type": "checkbox"},
    "_creator": {"column_type": "creator"},
    "_ctime": {"column_type": "ctime"},
    "_mtime": {"column_type": "mtime"},
    "_last_modifier": {"column_type": "last-modifier"},
}

#   dfdf



from ..column import (
    Text,
    LongText,
    Integer,
    Number,
    Collaborator,
    Date,
    Datetime,
    Duration,
    SingleSelect,
    MultipleSelect,
    Image,
    File,
    Email,
    Url,
    Checkbox,
    Rate,
    Formula,
    Link,
    LinkFomula,
    Creator,
    CreationTime,
    LastModifier,
    LastModificationTime,
    AutoNumber,
    Button,
)



from enum import Enum
from typing import List

from dataclasses_avroschema.avrodantic import AvroBaseModel
from pydantic import Extra, Field

__all__ = [
    "Text",
    "LongText",
    "Integer",
    "Number",
    "Collaborator",
    "Date",
    "Datetime",
    "Duration",
    "SingleSelect",
    "MultipleSelect",
    "Image",
    "File",
    "Email",
    "Url",
    "Checkbox",
    "Rate",
    "Formula",
    "Link",
    "LinkFomula",
    "Creator",
    "CreationTime",
    "LastModifier",
    "LastModificationTime",
    "AutoNumber",
    "Button",
    # Options
    "SelectOption",
    "OpenUrl",
]


# SeaTableBaseModel
class SeaTableType(AvroBaseModel):
    name: str
    anchor: str = None
    _type: str = "text"

    def seatable_schema(self):
        DEFAULT_FIELDS = ["name", "anchor"]
        schema = {
            "column_name": self.name,
            "column_type": self._type,
            "colum_data": dict(),
            "anchor_column": self.anchor,
        }

        for k in self.schema()["properties"]:
            if k not in DEFAULT_FIELDS:
                value = getattr(self, k)
                if value is None:
                    continue
                if "column_data" not in schema:
                    schema["column_data"] = dict()
                if isinstance(value, list):
                    value = [x.to_dict() if isinstance(x, AvroBaseModel) else x for x in value]
                else:
                    value = value.to_dict() if isinstance(value, AvroBaseModel) else value
                schema["column_data"].update({k: value})

        return {k: v for k, v in schema.items() if v}

    class Config:
        extra = Extra.forbid


# [TEXT]
class Text(SeaTableType):
    _type: str = "text"


# [LONG TEXT]
class LongText(SeaTableType):
    _type: str = "long-text"


# [NUMBER]
# number format
class NumberFormat(str, Enum):
    number: str = "number"
    percent: str = "percent"
    dollar: str = "dollar"
    euro: str = "euro"
    yen: str = "yen"


# number decimal
class NumberDecimal(str, Enum):
    dot: str = "dot"
    # comma: str = "comma"


# number thousands
class NumberThousands(str, Enum):
    no: str = "no"
    space: str = "space"
    comma: str = "comma"


# INTEGER
class Integer(SeaTableType):
    _type: str = "number"
    format: NumberFormat = "number"
    decimal: NumberDecimal = "dot"
    thousands: NumberThousands = "no"
    enable_precision: bool = True
    precision: int = 0


# NUMBER
class Number(SeaTableType):
    _type: str = "number"
    format: NumberFormat = "number"
    decimal: NumberDecimal = "dot"
    thousands: NumberThousands = "no"


# [COLLABORATOR]
class Collaborator(SeaTableType):
    _type: str = "collaborator"


# [DATE]
# DATE
class Date(SeaTableType):
    _type: str = "date"
    format: str = Field("YYYY-MM-DD", const=True)


# DATETIME
class Datetime(SeaTableType):
    _type: str = "date"
    format: str = Field("YYYY-MM-DD HH:mm", const=True)


# [DURATION]
class Duration(SeaTableType):
    _type: str = "duration"
    format: str = "duration"
    duration_format: str = "h:mm:ss"


# [SINGLE-SELECT, MULTIPLE-SELECT]
# select option
class SelectOption(SeaTableType):
    id: str
    color: str = None
    text_color: str = Field(None, alias="text-color")


# SINGLE SELECT
class SingleSelect(SeaTableType):
    _type = "single-select"
    options: List[SelectOption] = None


# MULTIPLE SELECT
class MultipleSelect(SeaTableType):
    _type = "multiple-select"
    options: List[SelectOption] = None


# [IMAGE]
class Image(SeaTableType):
    _type = "image"


# [FILE]
class File(SeaTableType):
    _type = "file"


# [EMAIL]
class Email(SeaTableType):
    _type = "email"


# [URL]
class Url(SeaTableType):
    _type = "url"


# [CHECKBOX]
class Checkbox(SeaTableType):
    _type = "checkbox"


# [Rate]
class Rate(SeaTableType):
    _type = "rate"
    rate_max_number: str = 5
    rate_style_color: str = "#FF8000"
    rate_style_type: str = "dtable-icon-rate"


# [FORMULA]
class Formula(SeaTableType):
    _type = "formula"
    formula: str = None


# [LINK COLUMN] - NOT WORKING!
class Link(SeaTableType):
    _type = "link"
    table: str
    other_table: str


# [LINK FORMULA COLUMN] - NOT WORKING!
class LinkFomula(SeaTableType):
    _type = "link-fomula"
    # [TBD]


# [CRAETOR]
class Creator(SeaTableType):
    _type = "creator"


# [CREATION TIME]
class CreationTime(SeaTableType):
    _type = "ctime"


# [LAST MODIFIER]
class LastModifier(SeaTableType):
    _type = "last-modifier"


# [LAST MODIFICATION TIME]
class LastModificationTime(SeaTableType):
    _type = "mtime"


# [AUTO NUMBER]
class AutoNumber(SeaTableType):
    _type = "auto-number"
    format: str = "ID-000000"


# [BUTTON]
# button type
class ButtonActionType(str, Enum):
    run_script: str = "run_script"
    send_email: str = "send_email"
    copy_row_to_another_table: str = "copy_row_to_another_table"
    modify_row: str = "modify_row"
    open_url: str = "open_url"


class ButtonAction(AvroBaseModel):
    action_type: ButtonActionType
    enable_condition_execution: bool = False


class Filter(AvroBaseModel):
    column_key: str
    filter_predicate: str = "is_not_empty"
    filter_term: str = ""


class OpenUrl(ButtonAction):
    action_type: ButtonActionType = Field("open_url", const=True)
    url_address: str
    enable_condition_execution: bool = False
    filters: List[Filter] = None


# Button
class Button(SeaTableType):
    _type = "button"
    button_name: str
    button_color: str
    button_action_list: List[ButtonAction] = None



from typing import List

from pydantic import BaseModel, root_validator, Field

from .const import (
    OP_DELETE_VIEW,
    OP_INSERT_VIEW,
    OP_MODIFY_VIEW_TYPE,
    OP_MODIFY_VIEW_LOCK,
    OP_MODIFY_FILTERS,
    OP_MODIFY_GROUPBYS,
    OP_MODIFY_HIDDEN_COLUMNS,
    OP_MODIFY_ROW_COLOR,
    OP_MODIFY_ROW_HEIGHT,
    OP_MODIFY_SORTS,
    OP_RENAME_VIEW,
    OP_MOVE_VIEW,
)
from .model import Event, View, Sort, Filter, Groupby


################################################################
# INSERT VIEW
################################################################
# View Event
class ViewEvent(Event):
    op_type: str
    table_id: str
    view_id: str


# Insert View
class InsertView(ViewEvent):
    view_data: dict
    view_folder_id: str = None

    @root_validator(pre=True)
    def update_view_id_from_table_data(cls, values):
        values.update({"view_id": values["view_data"]["_id"]})
        return values


# Delete View
class DeleteView(ViewEvent):
    view_folder_id: str = None
    view_name: str = None


# Rename View
class RenameView(ViewEvent):
    view_name: str


# Move View
class MoveView(ViewEvent):
    view_id: str = Field(alias="moved_view_id")
    target_view_id: str
    source_view_folder_id: str = None
    target_view_folder_id: str = None
    move_position: str
    moved_view_name: str


# Modify View Type
class ModifyViewType(ViewEvent):
    view_type: str


# Modify View Lock
class ModifyViewLock(ViewEvent):
    is_locked: bool


# Modify Filters
class ModifyFilters(ViewEvent):
    filters: List[Filter]
    filter_conjunction: str


# Modify Sorts
class ModifySorts(ViewEvent):
    sorts: List[Sort]


# Modify Groupbys
class ModifyGroupbys(ViewEvent):
    groupbys: List[Groupby]


# Modify Hidden COlumns
class ModifyHiddenColumns(ViewEvent):
    hidden_columns: List[str]


# Modify Row Color
class ModifyRowColor(ViewEvent):
    colorbys: dict


# Modify Row Height
class ModifyRowHeight(ViewEvent):
    row_height: str


################################################################
# View Event Parser
################################################################
def view_event_parser(data):
    op_type = data["op_type"]

    if op_type == OP_INSERT_VIEW:
        return [InsertView(**data)]

    if op_type == OP_DELETE_VIEW:
        return [DeleteView(**data)]

    if op_type == OP_RENAME_VIEW:
        return [RenameView(**data)]

    if op_type == OP_MODIFY_VIEW_TYPE:
        return [ModifyViewType(**data)]

    if op_type == OP_MODIFY_VIEW_LOCK:
        return [ModifyViewLock(**data)]

    if op_type == OP_MODIFY_FILTERS:
        return [ModifyFilters(**data)]

    if op_type == OP_MODIFY_SORTS:
        return [ModifySorts(**data)]

    if op_type == OP_MODIFY_GROUPBYS:
        return [ModifyGroupbys(**data)]

    if op_type == OP_MODIFY_HIDDEN_COLUMNS:
        return [ModifyHiddenColumns(**data)]

    if op_type == OP_MODIFY_ROW_COLOR:
        return [ModifyRowColor(**data)]

    if op_type == OP_MODIFY_ROW_HEIGHT:
        return [ModifyRowHeight(**data)]

    if op_type == OP_MOVE_VIEW:
        return [MoveView(**data)]

    _msg = f"view_parser - unknown op_type '{op_type}'."
    raise KeyError(_msg)



from .columns import column_event_parser
from .const import COLUMN_EVENTS, ROW_EVENTS, TABLE_EVENTS, VIEW_EVENTS
from .rows import row_event_parser
from .table import table_event_parser
from .view import view_event_parser


def event_parser(data):
    op_type = data["op_type"]

    if op_type in COLUMN_EVENTS:
        return column_event_parser(data)

    if op_type in ROW_EVENTS:
        return row_event_parser(data)

    if op_type in TABLE_EVENTS:
        return table_event_parser(data)

    if op_type in VIEW_EVENTS:
        return view_event_parser(data)

    _msg = f"event_parser - no parser for op_type '{op_type}'!"
    raise KeyError(_msg)



from typing import List

from pydantic import BaseModel, root_validator

from .const import (
    OP_DELETE_COLUMN,
    OP_INSERT_COLUMN,
    OP_MODIFY_COLUMN_METADATA_PERMISSION,
    OP_MODIFY_COLUMN_PERMISSION,
    OP_MODIFY_COLUMN_TYPE,
    OP_RENAME_COLUMN,
    OP_UPDATE_COLUMN_COLORBYS,
    OP_UPDATE_COLUMN_DESCRIPTION,
)
from .model import Column, ColumnMetadataPermission, ColumnPermission, Event


################################################################
# Models
################################################################
# Column Event
class ColumnEvent(Event):
    op_type: str
    table_id: str
    column_key: str


# Insert Column
class InsertColumn(ColumnEvent):
    anchor_column_key: str
    column_data: Column
    view_id: str
    rows_datas: list

    @root_validator(pre=True)
    def adjust_column_key(cls, values):
        values["anchor_column_key"] = values["column_key"]
        values.update({"column_key": values["column_data"]["key"]})
        return values


# Delete Column
class DeleteColumn(ColumnEvent):
    old_column: Column
    upper_column_key: str


# Rename Column
class RenameColumn(ColumnEvent):
    new_column_name: str
    old_column_name: str


# Update Column Description
class UpdateColumnDescription(ColumnEvent):
    column_description: str


# Modify Column Type
class ModifyColumnType(ColumnEvent):
    new_column: Column
    old_column: Column
    new_rows_data: List[dict]
    old_rows_data: List[dict]


# Modify Column Permission
class ModifyColumnPermission(ColumnEvent):
    new_column_permission: ColumnPermission
    old_column_permission: ColumnPermission


# Modify Column Metadata Permission
class ModifyColumnMetadataPermission(ColumnEvent):
    new_column_permission: ColumnMetadataPermission
    old_column_permission: ColumnMetadataPermission


################################################################
# COLUMN PARSER
################################################################
def column_event_parser(data):
    op_type = data["op_type"]

    if op_type == OP_INSERT_COLUMN:
        return [InsertColumn(**data)]

    if op_type == OP_DELETE_COLUMN:
        return [DeleteColumn(**data)]

    if op_type == OP_RENAME_COLUMN:
        return [RenameColumn(**data)]

    if op_type == OP_UPDATE_COLUMN_DESCRIPTION:
        return [UpdateColumnDescription(**data)]

    if op_type == OP_UPDATE_COLUMN_COLORBYS:
        pass

    if op_type == OP_MODIFY_COLUMN_TYPE:
        return [ModifyColumnType(**data)]

    if op_type == OP_MODIFY_COLUMN_PERMISSION:
        return [ModifyColumnPermission(**data)]

    if op_type == OP_MODIFY_COLUMN_METADATA_PERMISSION:
        return [ModifyColumnMetadataPermission(**data)]

    _msg = f"column parser - unknown op_type '{op_type}'!"
    raise KeyError(_msg)



import asyncio
import json
import logging
import time
import traceback
from typing import Callable, List

import redis.asyncio as redis

from ..conf import KEY_PREFIX, REDIS_HOST, REDIS_PORT

logger = logging.getLogger(__name__)


################################################################
# RedisStreamAdder
################################################################
class RedisStreamAdder:
    def __init__(
        self,
        host: str = REDIS_HOST,
        port: str = REDIS_PORT,
        key_prefix: str = KEY_PREFIX,
        maxlen: int = None,
        approximate: bool = False,
    ):
        # properties
        self.host = host
        self.port = port
        self.key_prefix = key_prefix
        self.maxlen = maxlen
        self.approximate = approximate

        # redis client
        self.redis_client = redis.Redis(host=self.host, port=self.port)

    async def __call__(self, key, msg):
        try:
            pipe = self.redis_client.pipeline()
            pipe.xadd("|".join([self.key_prefix, key]), msg)
            await pipe.execute()
        except Exception as ex:
            logger.warn(f"RedisStreamAdder {ex}")
            traceback.print_exc()



import logging

import click
import uvicorn
from click_loglevel import LogLevel

logger = logging.getLogger(__file__)


@click.group()
def plantable():
    pass


@plantable.command()
def hello():
    print("Hello, Plantable!")


@plantable.group()
def agent():
    pass


@agent.command()
@click.option("--seatable-url", default=None)
@click.option("--seatable-username", default=None)
@click.option("--seatable-password", default=None)
@click.option("--redis-host", default="localhost")
@click.option("--redis-port", default=6379)
@click.option("--log-level", default=logging.WARNING, type=LogLevel())
def run_producer(seatable_url, seatable_username, seatable_password, redis_host, redis_port, log_level):
    import asyncio

    from redis.exceptions import ConnectionError as RedisConnectionError

    from plantable.agent import Producer, RedisStreamAdder
    from plantable.agent.conf import (
        AWS_S3_ACCESS_KEY_ID,
        AWS_S3_SECRET_ACCESS_KEY,
        REDIS_HOST,
        REDIS_PORT,
        SEATABLE_PASSWORD,
        SEATABLE_URL,
        SEATABLE_USERNAME,
    )

    logging.basicConfig(level=log_level)

    REDIS_CONF = {
        "host": redis_host or REDIS_HOST,
        "port": redis_port or REDIS_PORT,
    }

    SEATABLE_CONF = {
        "seatable_url": seatable_url or SEATABLE_URL,
        "seatable_username": seatable_username or SEATABLE_USERNAME,
        "seatable_password": seatable_password or SEATABLE_PASSWORD,
    }

    async def main():
        handler = RedisStreamAdder(**REDIS_CONF)
        for _ in range(12):
            try:
                await handler.redis_client.ping()
                break
            except RedisConnectionError:
                print("Wait Redis...")
            await asyncio.sleep(5.0)

        producer = Producer(**SEATABLE_CONF, handler=handler)

        try:
            await producer.run()
        except asyncio.CancelledError:
            return

    asyncio.run(main())


@plantable.group()
def server():
    pass


@server.command()
@click.option("-h", "--host", type=str, default="0.0.0.0")
@click.option("-p", "--port", type=int, default=3000)
@click.option("--reload", is_flag=True)
@click.option("--workers", type=int, default=None)
@click.option("--log-level", type=LogLevel(), default=logging.INFO)
def run(host, port, reload, workers, log_level):
    logging.basicConfig(level=log_level)

    if reload:
        app = "plantable.server.app:app"
    else:
        from .server.app import app

    uvicorn.run(
        app,
        host=host,
        port=port,
        reload=reload,
        workers=workers,
        log_level=log_level,
    )
